# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""Public API for Fortran parser.

Module content
--------------
"""
# Author: Pearu Peterson <pearu@cens.ioc.ee>
# Created: Oct 2006


from six import string_types
# import all Statement classes:
from fparser.common.base_classes import classes
from fparser.common.utils import AnalyzeError

__autodoc__ = ['get_reader', 'parse', 'walk']


def get_reader(source, isfree=None, isstrict=None, include_dirs=None,
               source_only=None, ignore_comments=True):
    '''
    Returns Fortran reader instance.

    If ``source`` is a C filename then the functions searches for comment
    lines starting with ``/*f2py`` and reads following lines as PYF file
    content until a line ``*/`` is found.

    :param str source: Specify a string or filename containing Fortran code.
    :param bool isfree: True if Fortran is free format
    :param bool isstrict: True if we are to strictly enforce free/fixed format
    :param list include_dirs: Specify a list of include directories. The
                              default list (when include_dirs=None) contains
                              the current working directory and the directory
                              of ``source``.
    :param list source_only: Specify a list of Fortran file names that are
                             searched when the ``USE`` statement is
                             encountered.
    :param bool ignore_comments: Whether or not to ignore (and discard)
                                 comments when parsing the source.

    :returns: a reader instance
    :rtype: :py:class:`fparser.common.readfortran.FortranReader`
    '''
    import os
    import re
    from fparser.common.readfortran import FortranFileReader, \
        FortranStringReader
    from fparser.common.sourceinfo import FortranFormat

    if os.path.isfile(source):
        _name, ext = os.path.splitext(source)
        if ext.lower() in ['.c']:
            # get signatures from C file comments starting with
            # `/*f2py` and ending with `*/`.
            # TODO: improve parser to take line number offset making line
            #       numbers in parser messages correct.
            f2py_c_comments = re.compile(r'/[*]\s*f2py\s.*[*]/', re.I | re.M)
            handle = open(source, 'r')
            c_input = ''
            for line in f2py_c_comments.findall(handle.read()):
                c_input += line[2:-2].lstrip()[4:] + '\n'
            handle.close()
            if isfree is None:
                isfree = True
            if isstrict is None:
                isstrict = True
            return parse(c_input, isfree, isstrict, include_dirs)
        reader = FortranFileReader(source, include_dirs=include_dirs,
                                   source_only=source_only,
                                   ignore_comments=ignore_comments)
    elif isinstance(source, string_types):
        reader = FortranStringReader(source, include_dirs=include_dirs,
                                     source_only=source_only,
                                     ignore_comments=ignore_comments)
    else:
        raise TypeError('Expected string or filename input but got %s' %
                        (type(input)))
    if isfree is None:
        isfree = reader.format.is_free
    if isstrict is None:
        isstrict = reader.format.is_strict
    reader.set_format(FortranFormat(isfree, isstrict))
    return reader


def parse(source, isfree=None, isstrict=None, include_dirs=None,
          source_only=None, ignore_comments=True, analyze=True,
          clear_cache=True):
    """
    Parse input and return Statement tree. Raises an AnalyzeError if the
    parser can not parse the Fortran code.

    :param str source: Specify a string or filename containing Fortran code.
    :param bool isfree: Whether the Fortran source is free-format.
    :param bool isstrict: Whether we are to strictly enforce the `isfree`
                          setting.
    :param list include_dirs: Specify a list of include directories. The
                              default list (when include_dirs=None) contains
                              the current working directory and the directory
                              of ``filename``.
    :param list source_only: A list of Fortran file names that are searched
                             when the ``USE`` statement is encountered.
    :param bool ignore_comments: When True then discard all comment lines in
                                 the Fortran code.
    :param bool analyze: When True then apply analyze() method on the Fortran
                         code tree.
    :param bool clear_cache: Whether or not to wipe the parser cache prior
                             to parsing. Necessary when a new tree object
                             is required, even if the Fortran to be parsed has
                             been seen before.

    :returns: Abstract Syntax Tree of Fortran source.
    :rtype: :py:class:`fparser.api.BeginSource`
    """
    from fparser.one.parsefortran import FortranParser

    if clear_cache:
        # Wipe the parser cache if requested
        FortranParser.cache.clear()

    reader = get_reader(source, isfree, isstrict, include_dirs, source_only,
                        ignore_comments=ignore_comments)
    parser = FortranParser(reader, ignore_comments=ignore_comments)
    try:
        parser.parse()
    except AnalyzeError:
        raise
    if analyze:
        parser.analyze()

    return parser.block


def walk(stmt, depth=-1, _initial_depth=None):
    """ Generate Fortran statements by walking the stmt tree until given depth.

    For each block statement in stmt, the walk functions yields a
    tuple ``(statement, depth)`` where ``depth`` is the depth of tree
    stucture for statement.

    Parameters
    ----------
    stmt : Statement
    depth : int
      If depth is positive then walk in the tree until given depth.
      If depth is negative then walk the whole tree.

    Returns
    -------
    generator

    Examples
    --------

      ::

        from fparser import api
        source_str = '''
        subroutine foo
          integer i, r
          do i=1,100
            r = r + i
          end do
        end
        '''
        tree = api.parse(source_str)
        for stmt, depth in api.walk(tree):
            print depth, stmt.item

      that will print::

        1 line #2'subroutine foo'
        2 line #3'integer i, r'
        2 line #4'do i=1,100'
        3 line #5'r = r + i'
        2 line #6'end do'
        1 line #7'end'

    """
    if _initial_depth is None:
        if depth == 0:
            return
        _initial_depth = depth
    if not isinstance(stmt, classes.BeginSource):
        yield stmt, _initial_depth - depth
    if isinstance(stmt, classes.BeginStatement):
        last_stmt = stmt.content[-1]
        last_index = len(stmt.content)
        if isinstance(last_stmt, classes.EndStatement):
            last_index -= 1
        else:
            last_stmt = None
        if depth != 0:
            for substmt in stmt.content[:last_index]:
                for statement, statement_depth in walk(substmt, depth-1,
                                                       _initial_depth):
                    yield statement, statement_depth
        if last_stmt is not None:
            yield last_stmt, _initial_depth - depth



# Copyright (c) 2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""C99 Preprocessor Syntax Rules.
"""
# Author: Balthasar Reuter <balthasar.reuter@ecmwf.int>
# Based on previous work by Martin Schlipf (https://github.com/martin-schlipf)
# First version created: Jan 2020

import re
import sys

from fparser.common.readfortran import FortranReaderBase, CppDirective
from fparser.two import pattern_tools as pattern
from fparser.two.utils import (Base, StringBase, WORDClsBase)


# The list of classes that implement preprocessor directives
# This list is used in match_cpp_directive() and
# fparser.two.utils.BlockBase.match().
CPP_CLASS_NAMES = ['Cpp_If_Stmt', 'Cpp_Elif_Stmt', 'Cpp_Else_Stmt',
                   'Cpp_Endif_Stmt', 'Cpp_Include_Stmt', 'Cpp_Macro_Stmt',
                   'Cpp_Undef_Stmt', 'Cpp_Line_Stmt', 'Cpp_Error_Stmt',
                   'Cpp_Warning_Stmt', 'Cpp_Null_Stmt']


def match_cpp_directive(reader):
    '''Creates single-line C99 preprocessor directive object from the
    current line, if any is found.

    :param reader: the fortran file reader containing the line \
                   of code that we are trying to match.
    :type reader: :py:class:`fparser.common.readfortran.FortranFileReader` \
                  or \
                  :py:class:`fparser.common.readfortran.FortranStringReader`

    :return: the matched preprocessor directive object or `None`.
    :rtype: one of (:py:class:`fparser.two.C99Preprocess.Cpp_*_Stmt`,) or \
            `NoneType`

    '''
    # Assume we have potentially a CPP directive and only check for an
    # instance of CppDirective if reader is a subclass of FortranReaderBase
    is_potential_cpp_directive = True
    if isinstance(reader, FortranReaderBase):
        item = reader.get_item()
        is_potential_cpp_directive = isinstance(item, CppDirective)
        reader.put_item(item)
        # Do not bail out early here to have a catch all return statement
        # at the end (that would not be reachable by tests otherwise)
    if is_potential_cpp_directive:
        for cls in CPP_CLASS_NAMES:
            obj = getattr(sys.modules[__name__], cls)(reader)
            if obj:
                return obj
    return None


#
# ISO/IEC 9899: 1999 (C99)
#

#
# 6.10 Preprocessing directives
#

class Cpp_Pp_Tokens(Base):
    '''Generic class for preprocessor tokens that form the right hand
    side of a preprocessor directive (such as #error, #line, #if, etc.).
    '''

    subclass_names = []

    @staticmethod
    def match(string):
        '''Implements the matching for arbitrary preprocessor tokens
        that form the right hand side of a preprocessor directive. It
        does not impose any restrictions other than the string not
        being empty.

        :param str string: the string to be matched as pp-tokens.

        :return: a 1-tuple containing the matched string or None.
        :rtype: (str,) or NoneType

        '''
        if not string:
            return None
        line = string.strip()
        if not line:
            return None
        return (line,)

    def tostr(self):
        '''
        :return: this pp-tokens as a string.
        :rtype: str
        '''
        return self.items[0]


# 6.10.1 Conditional inclusion
# Deviating from the standard's definition, not entire if-else-endif
# blocks are treated as constructs but instead the relevant preprocessor
# directives are kept as single line nodes.


class Cpp_If_Stmt(WORDClsBase):
    '''
    C99 6.10.1 Conditional inclusion

    if-stmt is  # if constant-expression new-line
                or ifdef identifier new-line
                or ifndef identifier new-line

    '''

    subclass_names = []
    use_names = ['Cpp_Macro_Identifier', 'Cpp_Pp_Tokens']

    _regex = re.compile(r"#\s*(ifdef|ifndef|if)\b")
    _if_pattern = pattern.Pattern('<if>', r'^\s*#\s*if\b', value='#if')
    _def_pattern = (
        pattern.Pattern('<ifdef>', r'^\s*#\s*ifdef\b', value='#ifdef'),
        pattern.Pattern('<ifndef>', r'^\s*#\s*ifndef\b', value='#ifndef')
    )

    @staticmethod
    def match(string):
        '''Implements the matching for an if preprocessor directive
        (or its variations ifdef, ifndef). For ifdef and ifndef
        statements it matches the macro identifier using
        :py:class:`fparser.two.C99Preprocesser.Cpp_Macro_Identifier`
        otherwise it uses :py:class:`fparser.two.C99Preprocessor.Cpp_Pp_Tokens`
        to accept any non-empty string as rhs.

        :param str string: the string to match with as an if statement.

        :return: a tuple of size 2 containing the statement's keyword \
                  and the right hand side, or `None` if there is no match.
        :rtype: \
          (`str`, py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier`)\
          or (`str`, py:class:`fparser.two.C99Preprocessor.Cpp_Pp_Tokens`) \
          or `NoneType`

        '''
        if not string:
            return None
        result = WORDClsBase.match(
            Cpp_If_Stmt._if_pattern, Cpp_Pp_Tokens, string, colons=False,
            require_cls=True)
        return result or WORDClsBase.match(
            Cpp_If_Stmt._def_pattern, Cpp_Macro_Identifier, string,
            colons=False, require_cls=True)

    def tostr(self):
        '''
        :return: this if-stmt as a string.
        :rtype: str
        '''
        return '{0} {1}'.format(*self.items)


class Cpp_Elif_Stmt(WORDClsBase):
    '''
    C99 6.10.1 Conditional inclusion

    elif-stmt is  # elif constant-expression new-line

    '''

    subclass_names = []
    use_names = ['Cpp_Pp_Tokens']

    _pattern = pattern.Pattern('<elif-stmt>', r'^\s*#\s*elif\b', value='#elif')

    @staticmethod
    def match(string):
        '''Implements the matching for an elif preprocessor directive.

        :param str string: the string to match with as an elif statement.

        :return: a tuple of size 2 containing the statements keyword and \
                  right hand side, or `None` if there is no match.
        :rtype: \
            (`str`, :py:class:`fparser.two.C99_Preprocessor.Cpp_Pp_Tokens`) \
            or `NoneType`

        '''
        if not string:
            return None
        return WORDClsBase.match(
            Cpp_Elif_Stmt._pattern, Cpp_Pp_Tokens, string, colons=False,
            require_cls=True)

    def tostr(self):
        '''
        :return: this elif-stmt as a string.
        :rtype: str
        '''
        return '{0} {1}'.format(*self.items)


class Cpp_Else_Stmt(StringBase):
    '''
    C99 6.10.1 Conditional inclusion

    else-stmt is  # else new-line

    '''

    subclass_names = []

    _pattern = pattern.Pattern('<else>', r'^\s*#\s*else\b')

    @staticmethod
    def match(string):
        '''Implements the matching for an else preprocessor directive.

        :param str string: the string to match with as an else statement.

        :return: a 1-tuple containing the matched string or `None` if \
                  there is no match.
        :rtype: (str,) or NoneType

        '''
        if not string:
            return None
        return StringBase.match(Cpp_Else_Stmt._pattern, string)

    def tostr(self):
        '''
        :return: this else-stmt as a string.
        :rtype: str
        '''
        return self.string


class Cpp_Endif_Stmt(StringBase):
    '''
    C99 6.10.1 Conditional inclusion

    endif-stmt is  # endif new-line

    '''

    subclass_names = []

    _pattern = pattern.Pattern('<endif>', r'^\s*#\s*endif\b', value='#endif')

    @staticmethod
    def match(string):
        '''Implements the matching for an endif preprocessor directive.

        :param str string: the string to match with as an endif statement.

        :return: a 1-tuple containing the matched string or `None` if \
                  there is no match.
        :rtype: (str,) or NoneType

        '''
        if not string:
            return None
        return StringBase.match(Cpp_Endif_Stmt._pattern, string)

    def tostr(self):
        '''
        :return: this endif-stmt as a string.
        :rtype: str
        '''
        return self.string


class Cpp_Include_Stmt(Base):  # 6.10.2 Source file inclusion
    '''
    C99 6.10.2 Source file inclusion

    include_stmt is # include [ <h-char-sequence>
                               or "q-char-sequence"
                               or pp-tokens ] new-line
    '''

    _regex = re.compile(r"#\s*include\b")

    use_names = ['Include_Filename']

    @staticmethod
    def match(string):
        '''Implements the matching for an include statement.

        Allows for the filename to appear in double quotes or angle
        brackets. Only very loose restrictions are enforced for the
        filename, which is matched by
        py:class:`fparser.two.Fortran2003.Include_Filename`.

        :param str string: the string to match with as an include statement.

        :return: a tuple of size 1 containing a \
            py:class:`fparser.two.C99Preprocessor.Cpp_Include_Filename` \
            object with the matched filename if there is a match, \
            or `None` if there is not.
        :rtype: (:py:class:`fparser.two.Fortran2003.Include_Filename`, ) \
                or `NoneType`

        '''
        from fparser.two.Fortran2003 import Include_Filename
        if not string:
            return None
        line = string.strip()
        found = Cpp_Include_Stmt._regex.match(line)
        if not found:
            # The line does not match an include statement
            return None
        rhs = line[found.end():].strip()
        if rhs is None or len(rhs) < 3:
            # Either we didn't find any includes or the content after
            # the include token is too short to be valid (it must at
            # least contain quotes and one character.
            return None
        if not (rhs[0] == '"' and rhs[-1] == '"') and \
           not (rhs[0] == '<' and rhs[-1] == '>'):
            # The filename should be surrounded by double
            # quotes or '<...>' but this is not the case.
            return None
        # Remove the quotes.
        file_name = rhs[1:-1]
        # Pass the potential filename to the relevant class.
        return (Include_Filename(file_name),)

    def tostr(self):
        '''
        :return: this include-stmt as a string.
        :rtype: str
        '''
        return '#include "{0}"'.format(self.items[0])


class Cpp_Macro_Stmt(Base):  # 6.10.3 Macro replacement
    """
    C99 6.10.3 Macro replacement

    macro_stmt is # define identifier [( [identifier-list] ) or (...) ]
                  [ replacement-list ] new-line

    Important: No preceding whitespace is allowed for the left parenthesis of
    the optional identifier-list. If a preceding whitespace is encountered,
    the bracket is considered part of the replacement-list.
    """

    use_names = ['Cpp_Macro_Identifier', 'Cpp_Macro_Identifier_List',
                 'Cpp_Pp_Tokens']

    _regex = re.compile(r"#\s*define\b")

    @staticmethod
    def match(string):
        '''Implements the matching for a preprocessor macro definition.

        It matches define directives with macro identifier, optional
        identifier list, and optional replacement list. The macro
        identifier is matched using
        :py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier`
        and the optional argument identifier list using
        :py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier_List`.

        Important: No preceding whitespace is allowed for the left
        parentheses of the dentifier-list. If a preceding  whitespace is
        encountered, the it is considered part of the replacement-list.

        :param str string: the string to match with as an if statement.

        :return: a tuple of size 3 containing the macro identifier, \
                  identifier list or None, and replacement list or None, \
                  or `None` if there is no match.
        :rtype: \
          (py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier`, \
           py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier_List` \
           or NoneType, \
           py:class:`fparser.two.C99Preprocessor.Cpp_Pp_Tokens` or NoneType) \
          or `NoneType`

        '''
        if not string:
            return None
        line = string.strip()
        found = Cpp_Macro_Stmt._regex.match(line)
        if not found:
            # The line does not match a define statement
            return None
        rhs = line[found.end():].strip()
        found = pattern.macro_name.match(rhs)
        if not found:
            return None
        name = Cpp_Macro_Identifier(found.group())
        definition = rhs[found.end():]
        # note no strip here because '#define MACRO(x)' and
        # '#define MACRO (x)' are functionally different
        if not definition:
            return (name, None, None)
        if definition[0] == '(':
            found = Cpp_Macro_Identifier_List._pattern.match(definition)
            if not found:
                # The definition starts with a bracket (without preceding
                # white space) but does not match an identifier list
                return None
            parameter_list = Cpp_Macro_Identifier_List(found.group())
            # Note that the definition can potentially be an empty string
            # which is nonetheless included explicitly to distinguish a
            # macro definition with arguments from one without but with
            # a definition
            definition = definition[found.end():]
        else:
            parameter_list = None
        # now that definition only holds the replacement list, we can strip
        definition = definition.strip()
        if definition:
            definition = Cpp_Pp_Tokens(definition)
        else:
            definition = None
        return (name, parameter_list, definition)

    def tostr(self):
        '''
        :return: this macro-stmt as a string.
        :rtype: str
        '''
        return '#define {0}{1}{2}{3}'.format(
            self.items[0], self.items[1] or '', ' ' if self.items[2] else '',
            self.items[2] or '')


class Cpp_Macro_Identifier(StringBase):  # pylint: disable=invalid-name
    '''Implements the matching of a macro identifier.'''

    # There are no other classes. This is a simple string match.
    subclass_names = []

    @staticmethod
    def match(string):
        '''Implements the matching of a macro identifier.

        It matches the string with the regular expression abs_macro_name
        in the pattern_tools file. The macro identifier may contain
        only letters and underscore.

        :param str string: the string to match with the pattern rule.

        :return: a tuple of size 1 containing a string with the \
                 matched name if there is a match, or None if there is not.
        :rtype: (str)  or NoneType

        '''
        return StringBase.match(pattern.abs_macro_name, string.strip())


class Cpp_Macro_Identifier_List(StringBase):
    '''Implements the matching of an identifier list in a macro definition.

    identifier-list is (identifier [, identifier-list or ...])
                       or (...)
    '''

    subclass_names = []

    _pattern = pattern.Pattern('<identifier-list>',
                               r'\((\s*[A-Za-z_]\w*'
                               r'(?:\s*,\s*[A-Za-z_]\w*)*'
                               r'(?:\s*,\s*\.{3})?|\.{3})?\s*\)')

    @staticmethod
    def match(string):
        '''Implements the matching of a macro identifier list as part of
        a macro definition. It must consist of one or more macro
        identifier separated by comma, or "..." for a variadic argument
        list, and must be surrouned by parentheses.

        For simplicity, the matched list is kept as a single string and not
        matched as
        :py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier`.

        :param str string: the string to match with the pattern rule.

        :return: a tuple of size 1 containing a string with the \
                 matched identifier list if there is a match, or None if \
                 there is not.
        :rtype: (`str`,) or `NoneType`

        '''
        if not string:
            return None
        return StringBase.match(Cpp_Macro_Identifier_List._pattern, string)

    def tostr(self):
        '''
        :return: this macro-identifier-list as a string.
        :rtype: str
        '''
        return self.string


class Cpp_Undef_Stmt(WORDClsBase):
    '''Implements the matching of a preprocessor undef statement for a macro.

    undef-stmt is # undef identifier new-line

    Strictly, this is part of 6.10.3 but since it is identified by a different
    directive keyword (undef instead of define) we treat it separately.

    '''

    subclass_names = []

    use_names = ['Cpp_Macro_Identifier']

    _pattern = pattern.Pattern('<undef>', r'^\s*(#\s*undef)\b', value='#undef')

    @staticmethod
    def match(string):
        '''Implements the matching for a preprocessor undef statement
        for a macro. The macro identifier is matched using
        :py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier`.

        :param str string: the string to match with as an if statement.

        :return: a tuple of size 1 containing the macro identifier, or\
                  `None` if there is no match.
        :rtype: (py:class:`fparser.two.C99Preprocessor.Cpp_Macro_Identifier`) \
                or `NoneType`

        '''
        if not string:
            return None
        return WORDClsBase.match(
            Cpp_Undef_Stmt._pattern, Cpp_Macro_Identifier, string,
            colons=False, require_cls=True)

    def tostr(self):
        '''
        :return: this undef-stmt as a string.
        :rtype: str
        '''
        return '{0} {1}'.format(*self.items)


class Cpp_Line_Stmt(WORDClsBase):  # 6.10.4 Line control
    '''
    C99 6.10.4 Line control

    line-stmt is # line digit-sequence [ "s-char-sequence" ] new-line
                        or pp-tokens new-line
    '''

    subclass_names = []
    use_names = ['Cpp_Pp_Tokens']

    _pattern = pattern.Pattern('<line>', r'^\s*#\s*line\b', value='#line')

    @staticmethod
    def match(string):
        '''Implements the matching for a line preprocessor directive.
        The right hand side of the directive is not matched any further
        but simply kept as a string.

        :param str string: the string to match with as a line statement.

        :return: a tuple of size 1 with the right hand side as a string, \
                  or `None` if there is no match.
        :rtype: (`str`) or `NoneType`

        '''
        if not string:
            return None
        return WORDClsBase.match(
            Cpp_Line_Stmt._pattern, Cpp_Pp_Tokens, string, colons=False,
            require_cls=True)

    def tostr(self):
        '''
        :return: this line-stmt as a string.
        :rtype: str
        '''
        return '{0} {1}'.format(*self.items)


class Cpp_Error_Stmt(WORDClsBase):  # 6.10.5 Error directive
    '''
    C99 6.10.5 Error directive

    error-stmt is # error [pp-tokens] new-line

    '''

    subclass_names = []
    use_names = ['Cpp_Pp_Tokens']

    _pattern = pattern.Pattern('<error>', r'^\s*#\s*error\b', value='#error')

    @staticmethod
    def match(string):
        '''Implements the matching for an error preprocessor directive.
        The optional right hand side of the directive is not matched any
        further but simply kept as a string.

        :param str string: the string to match with as a line statement.

        :return: an empty tuple or a tuple of size  1 with the right hand \
                  side as a string, or `None` if there is no match.
        :rtype: () or (`str`) or `NoneType`

        '''
        if not string:
            return None
        return WORDClsBase.match(
            Cpp_Error_Stmt._pattern, Cpp_Pp_Tokens, string, colons=False,
            require_cls=False)

    def tostr(self):
        '''
        :return: this error-stmt as a string.
        :rtype: str
        '''
        if self.items[1]:
            return '{0} {1}'.format(*self.items)
        return self.items[0]


class Cpp_Warning_Stmt(WORDClsBase):
    '''
    Not actually part of C99 but supported by most preprocessors and
    with syntax identical to Cpp_Error_Stmt

    warning-stmt is # warning [pp-tokens] new-line

    '''

    subclass_names = []
    use_names = ['Cpp_Pp_Tokens']

    _pattern = pattern.Pattern('<warning>', r'^\s*#\s*warning\b',
                               value='#warning')

    @staticmethod
    def match(string):
        '''Implements the matching for a warning preprocessor directive.
        The optional right hand side of the directive is not matched any
        further but simply kept as a string.

        :param str string: the string to match with as a line statement.

        :return: an empty tuple or a tuple of size  1 with the right hand \
                  side as a string, or `None` if there is no match.
        :rtype: () or (`str`) or `NoneType`

        '''
        if not string:
            return None
        return WORDClsBase.match(
            Cpp_Warning_Stmt._pattern, Cpp_Pp_Tokens, string, colons=False,
            require_cls=False)

    def tostr(self):
        '''
        :return: this warning-stmt as a string.
        :rtype: str
        '''
        if self.items[1]:
            return '{0} {1}'.format(*self.items)
        return self.items[0]


# 6.10.6 Pragma directive
# Pragma Preprocessor directives not implemented since Fortran has its own
# Pragma syntax in the form of comments. For that reason, most preprocessors
# do not support C preprocess pragmas in Fortran code either.


class Cpp_Null_Stmt(Base):  # 6.10.7 Null directive
    '''
    C99 6.10.7  Null directive

    null-stmt is # new-line

    '''

    subclass_names = []

    @staticmethod
    def match(string):
        '''Implements the matching for a Null (empty) directive.

        :param str string: the string to match with as a line statement.

        :return: an empty tuple or `None` if there is no match.
        :rtype: () or `NoneType`

        '''
        if not string:
            return None
        line = string.strip()
        if not line == '#':
            return None
        return ()

    def tostr(self):
        '''
        :return: this null-stmt as a string.
        :rtype: str
        '''
        return '#'



# Copyright (c) 2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R603 : This file tests support for the
Designator class.

Three of the rules in Designator class are currently not called as the
text matches to earlier rules. The associated rule constraints and
specification text (if implemented, which they are not - see issue
#201) would fix this problem.

For example, here is note 6.6 from the F2003 specification which helps
clarify for derived types:

The syntax rules are structured such that a data-ref that ends in a
component name without a following subscript list is a structure
component, even when other component names in the dataref are followed
by a subscript list. A data-ref that ends in a component name with a
following subscript list is either an array element or an array
section. A data-ref of nonzero rank that ends with a substring-range
is an array section. A data-ref of zero rank that ends with a
substring-range is a substring.

'''
import pytest
from fparser.two.Fortran2003 import Designator, Name, Part_Ref, \
    Array_Section, Program, Data_Ref, NoMatchError, \
    Section_Subscript_List, Array_Element
from fparser.api import get_reader


@pytest.mark.usefixtures("f2003_create")
def test_object_name():
    '''Test that a Name object is returned when a valid object-name name
    is supplied

    '''
    result = Designator("fred")
    assert str(result) == "fred"
    assert isinstance(result, Name)


@pytest.mark.usefixtures("f2003_create")
def test_array_element():
    '''Test that the array-element rule is matched when the Designator
    rule is supplied with an array-element.

    # Note that at this time this example would also match rules
    # array-section and structure-component (although it would give
    # the same result so would not matter). It matches the expected
    # rule (array-element) as this rule occurs before array-section
    # and structure-component in the Designator class subclass_names
    # list so is matched first.

    '''
    # A Part_Ref is returned (as array-element -> data-ref -> part-ref
    # and a data-ref is not returned if there is only one part-ref)
    result = Designator("a(7,b(i)+c)")
    assert str(result) == "a(7, b(i) + c)"
    assert isinstance(result, Part_Ref)

    # A Data_Ref containing a Part_Ref is returned (as array-element
    # -> data-ref -> part-ref and a data-ref is returned is there is
    # more than one part-ref)
    result = Designator("x%a(1)")
    assert str(result) == "x % a(1)"
    assert isinstance(result, Data_Ref)
    assert isinstance(result.items[0], Name)  # x
    assert isinstance(result.items[1], Part_Ref)  # a(1)


@pytest.mark.xfail(reason="issue #201 Array section is parsed as an array "
                   "element.")
@pytest.mark.usefixtures("f2003_create")
def test_array_section():
    '''Test that the array-section rule is matched when the Designator
    rule is supplied with an array-section.

    '''
    result = Designator("a(1:2)")
    assert str(result) == "a(1 : 2)"
    assert isinstance(result, Part_Ref)
    assert isinstance(result.items[0], Name)  # a
    assert isinstance(result.items[1], Section_Subscript_List)  # 1:2

    # This example should not match Array_Element as, if it does, it
    # will follow the wrong rule hierarchy (as array-element occurs
    # before array-section in the Designator class
    # subclass_names). This would be avoided if the appropriate
    # constraints were checked, but they are not. However, this does
    # not really matter in this example as we end up with the same
    # parse tree hierarchy in any case (as both rule hierarchies end
    # up with a match to part_ref).
    with pytest.raises(NoMatchError):
        _ = Array_Element("a(1:2)")


@pytest.mark.usefixtures("f2003_create")
def test_array_section2():
    '''Test that an Array_Section object is returned when a valid
    array-section (followed by a substring-range) is supplied.

    '''
    result = Designator("a(1:2:3)(:)")
    assert str(result) == "a(1 : 2 : 3)(:)"
    assert isinstance(result, Array_Section)


@pytest.mark.xfail(reason="issue #201 Structure-component is parsed as an "
                   "array element.")
@pytest.mark.usefixtures("f2003_create")
def test_structure_component():
    '''Test that the structure-component rule is matched when the Designator
    rule is supplied with a structure component.

    '''
    result = Designator("parent%scalar_field")
    assert str(result) == "parent % scalar_field"
    assert isinstance(result, Data_Ref)

    # This example should not match Array_Element or Array_Section as,
    # if it does, it will follow the wrong rule hierarchy (as
    # array-element and array-section occur before structure-component
    # in the Designator class subclass_names). This would be avoided
    # if the appropriate constraints were checked, but they are
    # not. However, this does not really matter in this example as we
    # end up with the same parse tree hierarchy in any case (as the
    # rule hierarchies end up with a match to part_ref).
    with pytest.raises(NoMatchError):
        _ = Array_Element("parent%scalar_field")
    with pytest.raises(NoMatchError):
        _ = Array_Section("parent%scalar_field")


@pytest.mark.xfail(reason="issue #201 Substring is parsed as an array "
                   "element.")
@pytest.mark.usefixtures("f2003_create")
def test_substring():
    '''Test that an Substring object is returned when a valid substring is
    supplied. This example provides the context (i.e. 'a' is a
    character string) as otherwise it is not possible for the parser
    to distinguish between a substring and an array section and it
    defaults to the latter.

    '''
    reader = get_reader(
        "program test\n"
        "character(len=10) :: a\n"
        "a(1:3)='hey'\n"
        "end program test\n")
    result = Program(reader)
    assert "a(1 : 3) = 'hey'" in str(result)
    assert isinstance(result, Program)
    assert "Substring" in repr(result)



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R612 : This file tests support for the
Data_Ref class.

'''

import pytest
from fparser.two.utils import NoMatchError
from fparser.two.Fortran2003 import Data_Ref, Part_Ref, Name


def test_valid_sequence(f2003_create):
    '''Test that a data_ref object is returned when a valid sequence is
    supplied.

    '''
    for string in ["a%b", " a % b ", "a%b%c", "A%B%C"]:
        result = Data_Ref(string)
        assert str(result) == str(result).strip()
        assert isinstance(result, Data_Ref)


def test_single_entry(f2003_create):
    '''Test that a data_ref object is not returned when the sequence is
    valid but contains a single entry.

    '''
    for string in ["a", " a ", "A"]:
        result = Data_Ref(string)
        assert str(result) == str(result).strip()
        assert isinstance(result, Name)

    result = Data_Ref("a(2)")
    assert str(result) == str(result).strip()
    assert isinstance(result, Part_Ref)


def test_invalid(f2003_create):
    '''Test that there is no match when the input is invalid. '''

    for string in ["", "  ", "1", "%", "a b"]:
        with pytest.raises(NoMatchError):
            assert Data_Ref(string) is None



# -*- coding: utf-8 -*-
##############################################################################
# Copyright (c) 2017 Science and Technology Facilities Council
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##############################################################################
# Modified M.Hambley, UK Met Office
##############################################################################
'''
Test battery associated with fparser.common.base_classes package.
'''
import re
import pytest

import fparser.common.base_classes
import fparser.common.readfortran
import fparser.common.sourceinfo
import fparser.common.utils
from fparser import api


def test_statement_logging(log, monkeypatch):
    '''
    Tests the Statement class' logging methods.
    '''
    class DummyParser(object):
        '''
        Null parser harness.
        '''
        def __init__(self, reader):
            self.reader = reader

    reader = fparser.common.readfortran.FortranStringReader("dummy = 1")
    parser = DummyParser(reader)

    monkeypatch.setattr(fparser.common.base_classes.Statement,
                        'process_item', lambda x: None, raising=False)
    unit_under_test = fparser.common.base_classes.Statement(parser, None)

    unit_under_test.error('Scary biscuits')
    assert(log.messages == {'critical': [],
                            'debug':    [],
                            'error':    ['Scary biscuits'],
                            'info':     [],
                            'warning':  []})

    log.reset()
    unit_under_test.warning('Trepidacious Cetations')
    assert(log.messages == {'critical': [],
                            'debug':    [],
                            'error':    [],
                            'info':     [],
                            'warning':  ['Trepidacious Cetations']})

    log.reset()
    unit_under_test.info('Hilarious Ontologies')
    assert(log.messages == {'critical': [],
                            'debug':    [],
                            'error':    [],
                            'info':     ['Hilarious Ontologies'],
                            'warning':  []})


def test_log_comment_mix(log):
    '''
    Tests that unexpected Fortran 90 comment in fixed format source is logged.
    '''
    class EndDummy(fparser.common.base_classes.EndStatement):
        '''
        Dummy EndStatement.
        '''
        match = re.compile(r'\s*end(\s*thing\s*\w*|)\s*\Z', re.I).match

    class BeginHarness(fparser.common.base_classes.BeginStatement):
        '''
        Dummy BeginStatement.
        '''
        end_stmt_cls = EndDummy
        classes = []
        match = re.compile(r'\s*thing\s+(\w*)\s*\Z', re.I).match

        def get_classes(self):
            '''
            Returns an empty list of contained statements.
            '''
            return []

    code = '      x=1 ! Cheese'
    parent = fparser.common.readfortran.FortranStringReader(
        code, ignore_comments=False)
    parent.set_format(fparser.common.sourceinfo.FortranFormat(False, True))
    item = fparser.common.readfortran.Line(code, (1, 1), None, None, parent)
    with pytest.raises(fparser.common.utils.AnalyzeError):
        __ = BeginHarness(parent, item)
    expected = '    1:      x=1 ! Cheese <== ' \
               + 'no parse pattern found for "x=1 ! cheese" ' \
               + "in 'BeginHarness' block, " \
               + 'trying to remove inline comment (not in Fortran 77).'
    result = log.messages['warning'][0].split('\n')[1]
    assert result == expected


def test_log_unexpected(log):
    '''
    Tests that an unexpected thing between begin and end statements logs an
    event.
    '''
    class EndThing(fparser.common.base_classes.EndStatement):
        '''
        Dummy EndStatement class.
        '''
        isvalid = True
        match = re.compile(r'\s*end(\s+thing(\s+\w+)?)?\s*$', re.I).match

    class BeginThing(fparser.common.base_classes.BeginStatement):
        '''
        Dummy BeginStatement class.
        '''
        end_stmt_cls = EndThing
        classes = []
        match = re.compile(r'\s*thing\s+(\w+)?\s*$', re.I).match

        def get_classes(self):
            '''
            Returns an empty list of contained classes.
            '''
            return []

    code = ['      jumper', '      end thing']
    parent = fparser.common.readfortran.FortranStringReader('\n'.join(code))
    parent.set_format(fparser.common.sourceinfo.FortranFormat(False, True))
    item = fparser.common.readfortran.Line(code[0], (1, 1), None, None, parent)
    with pytest.raises(fparser.common.utils.AnalyzeError):
        __ = BeginThing(parent, item)
    expected = '    1:      jumper <== no parse pattern found for "jumper" ' \
               "in 'BeginThing' block."
    result = log.messages['warning'][0].split('\n')[1]
    assert result == expected


def test_space_after_enddo():
    '''Make sure that there is no space after an 'END DO' without name,
    but there is a space if there is a name after 'END DO'.
    '''

    # Unnamed loop:
    source_str = '''\
    subroutine foo
    integer i, r
    do i = 1,100
      r = r + 1
    end do
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert "END DO " not in tree.tofortran()

    # Named loop:
    source_str = '''\
    subroutine foo
    integer i, r
    loop1: do i = 1,100
      r = r + 1
    end do loop1
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert "END DO loop1" in tree.tofortran()



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Science and Technology Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
Tests the parser against "do" block syntax.

This is a complicated piece of syntax with many axes of movement. A
comprehensive testing was felt to be too onerous so this stripped down
version checks only a subset. Hopefully it is representative.
'''
from __future__ import absolute_import, print_function

import pytest

from fparser.common.utils import AnalyzeError
from fparser.common.sourceinfo import FortranFormat
from fparser.one.parsefortran import FortranParser
from fparser.common.readfortran import FortranStringReader


def get_end_do(name):
    '''A small helper function to return either "END DO" (without space) if
    name is empty, or "END DO "+name. This simplifies the tests now that
    tofortran does not return an "END DO" with one space in case of an
    unnamed end statement.

    :param str name: Either None if it is an unnamed statement, or \
        the label to use in the end statement.

    :returns: either "END DO" (without space) if name is empty, or \
        "END DO "+name.
    :rtype: str

    '''

    if name:
        return "END DO {0}".format(name)
    return "END DO"


@pytest.mark.parametrize('name', [None, 'loop_name'])
@pytest.mark.parametrize('label', [None, '123'])
@pytest.mark.parametrize('control_comma', [False, True])
@pytest.mark.parametrize('terminal_expression',
                         ['1', '10', 'x+y', 'size(array)', 'size(this%array)'])
@pytest.mark.parametrize('end_name', [None, 'loop_name', 'wrong_name'])
@pytest.mark.parametrize('end_label', [None, '123', '456'])
def test_do(name, label, control_comma, terminal_expression,
            end_name, end_label):
    # pylint: disable=redefined-outer-name, too-many-arguments, too-many-locals
    '''
    Checks that the "do" loop parser understands the "for-next" variant of
    the syntax. This is defined in BS ISO/IEC 1539-1:2010 with R814-R822.

    TODO: Only the terminal expression is tested. This is a short-cut and
          relies on expression handling being applied identically across
          all expressions. This was true at the time of writing the test.
    '''
    name_snippet = name + ': ' if name else None
    label_snippet = label + ' ' if label else None
    comma_snippet = ', ' if control_comma else None
    # TODO: Although the Fortran standard allows for "continue" to be used in
    # place of "end do" fparser does not support it.
    end_snippet = 'continue' if end_name == 'continue' \
                  else get_end_do(end_name)
    do_code = '''{name}do {label}{comma}variable = 1, {term}, 1
  write (6, '(I0)') variable
{endlabel} {end}
'''.format(name=name_snippet or '',
           label=label_snippet or '',
           comma=comma_snippet or '',
           term=terminal_expression,
           endlabel=end_label or '',
           end=end_snippet)
    do_expected = '''  {name}DO {label}variable = 1, {term}, 1
    WRITE (6, '(I0)') variable
{endlabel} {endstmt}
'''.format(name=name_snippet or '',
           label=label_snippet or '',
           term=terminal_expression,
           endlabel=end_label or ' ',
           endstmt=get_end_do(end_name))
    do_reader = FortranStringReader(do_code)
    do_reader.set_format(FortranFormat(True, False))
    do_parser = FortranParser(do_reader)
    if (name != end_name) or (label and (label != end_label)):
        with pytest.raises(AnalyzeError):
            do_parser.parse()
    else:
        do_parser.parse()
        loop = do_parser.block.content[0]
        assert str(loop).splitlines() == do_expected.splitlines()


@pytest.mark.parametrize('name', [None, 'loop_name'])
@pytest.mark.parametrize('label', [None, '123'])
@pytest.mark.parametrize('control_comma', [False, True])
@pytest.mark.parametrize('terminal_expression',
                         ['1', 'x+y', 'size(array)', 'size(this%array)'])
@pytest.mark.parametrize('end_name', [None, 'loop_name', 'wrong_name'])
@pytest.mark.parametrize('end_label', [None, '123', '456'])
def test_do_while(name, label, control_comma, terminal_expression,
                  end_name, end_label):
    # pylint: disable=redefined-outer-name, too-many-arguments
    '''
    Checks that the "do" loop parser understands the "do-while" variant of
    the syntax. This is defined in BS ISO/IEC 1539-1:2010 with R814-R822.
    '''
    name_snippet = name + ': ' if name else None
    label_snippet = label + ' ' if label else None
    comma_snippet = ', ' if control_comma else None
    code = '''{name}do {label}{comma}while ({term})
  write (6, '(I0)') variable
{endlabel} {endstmt}
'''.format(name=name_snippet or '',
           label=label_snippet or '',
           comma=comma_snippet or '',
           term=terminal_expression,
           endlabel=end_label or '',
           endstmt=get_end_do(end_name))
    expected = '''  {name}DO {label}while ({term})
    WRITE (6, '(I0)') variable
{endlabel} {endstmt}
'''.format(name=name_snippet or '',
           label=label_snippet or '',
           term=terminal_expression,
           endlabel=end_label or ' ',
           endstmt=get_end_do(end_name))
    print(code)
    reader = FortranStringReader(code)
    reader.set_format(FortranFormat(True, False))
    parser = FortranParser(reader)
    if (name != end_name) or (label and (label != end_label)):
        with pytest.raises(AnalyzeError):
            parser.parse()
    else:
        parser.parse()
        loop = parser.block.content[0]
        assert str(loop).splitlines() == expected.splitlines()



# Modified work Copyright (c) 2017-2020 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

'''Base classes and exception handling for Fortran parser

'''
# Original author: Pearu Peterson <pearu@cens.ioc.ee>
# First version created: Oct 2006

import re
import logging
import six
from fparser.common.splitline import string_replace_map
from fparser.two import pattern_tools as pattern
from fparser.common.readfortran import FortranReaderBase


# A list of supported extensions to the standard(s)

# An X edit descriptor in a format statement specifies the position
# (forward from the current position) at which the next character will
# be transmitted to or from a record. In standard Fortran2003 the X
# edit descriptor must be preceeded by an integer which specifies how
# far forward from the current position. The 'x-format' extension
# allows the X edit descriptor to be specified without a preceeding
# integer.
EXTENSIONS = ["x-format"]

# Cray pointers are a well known extension to the Fortran
# standard. See http://pubs.cray.com/content/S-3901/8.6/
# cray-fortran-reference-manual-s-3901-86/types or
# https://gcc.gnu.org/onlinedocs/gfortran/Cray-pointers.html for
# example. If 'cray-pointer' is specified in EXTENSIONS then this
# extension is allowed in fparser.
EXTENSIONS += ["cray-pointer"]

# A Hollerith constant is a way of specifying a string as a sequence
# of characters preceded by the string length and separated by an 'H'
# e.g. 5Hhello. See
# https://gcc.gnu.org/onlinedocs/gfortran/Hollerith-constants-support.html,
# for example for more details. fparser currently supports Hollerith
# constants specified in format statements when 'hollerith' is specified
# in the EXTENSIONS list.
EXTENSIONS += ["hollerith"]

# Many compilers support the use of '$' in a fortran write statement
# to indicate that the carriage return should be suppressed. This is
# an extension to the Fortran standard and is supported by fparser if
# 'dollar-descriptor' is specified in the EXTENSIONS list.
EXTENSIONS += ["dollar-descriptor"]


def py2_encode_list_items(mylist):
    '''
    If we are running under Python2 then ensure that all strings in the
    supplied list are encoded as bytes. If we are running under Python3
    then all strings are unicode and this routine does nothing.

    :param list mylist: List of strings to modify.

    '''
    if six.PY2:
        # If we're in Python2 then we have to take care as comments
        # and character literals may be UTF while other elements are not.
        for idx, item in enumerate(mylist[:]):
            if isinstance(item, six.text_type):
                # This item is UTF so encode it as bytes
                mylist[idx] = item.encode('utf-8')


class FparserException(Exception):
    '''Base class exception for fparser. This allows an external tool to
    capture all exceptions if required.

    :param str info: a string giving contextual error information.

    '''
    def __init__(self, info):
        Exception.__init__(self, info)


class NoMatchError(FparserException):
    '''An exception indicating that a particular rule implemented by a
    class does not match the provided string. It does not necessary
    mean there is an error as another rule may match. This exception
    is used internally so should never be visible externally.

    '''


class FortranSyntaxError(FparserException):
    '''An exception indicating that fparser believes the provided code to
    be invalid Fortran. Also returns information about the location of
    the error if that information is available.

    :param reader: input string or reader where the error took \
    place. This is used to provide line number and line content \
    information.
    :type reader: str or :py:class:`FortranReaderBase`
    :param str info: a string giving contextual error information.

    '''
    def __init__(self, reader, info):
        output = "at unknown location "
        if isinstance(reader, FortranReaderBase):
            output = "at line {0}\n>>>{1}\n".format(
                reader.linecount,
                reader.source_lines[reader.linecount-1])
        if info:
            output += "{0}".format(info)
        FparserException.__init__(self, output)


class InternalError(FparserException):
    '''An exception indicating that an unexpected error has occured in the
    parser.

    :param str info: a string giving contextual error information.

    '''
    def __init__(self, info):
        new_info = ("'{0}'. Please report this to the "
                    "authors.".format(info))
        FparserException.__init__(self, new_info)


class InternalSyntaxError(FparserException):
    '''An exception indicating that a syntax error has been found by the
    parser. This is used instead of `FortranSyntaxError` when the
    reader object is not available.

    '''


def show_result(func):
    return func

    def new_func(cls, string, **kws):
        r = func(cls, string, **kws)
        if r is not None and isinstance(r, StmtBase):
            print('%s(%r) -> %r' % (cls.__name__, string, str(r)))
        return r
    return new_func


#
# BASE CLASSES
#


class ComparableMixin(object):
    """ Mixin class to provide rich comparison operators.

    This mixin provides a set of rich comparison operators. Each class using
    this mixin has to provide a _cmpkey() method that returns a key of objects
    that can be compared.

    See also http://python3porting.com/preparing.html#richcomparisons
    """
    # pylint: disable=too-few-public-methods

    def _compare(self, other, method):
        """ Call the method, if other is able to be used within it.

        :param object other: The other object to compare with
        :type other: object
        :param method: The method to call to compare self and other.
        :type method: LambdaType
        :return: NotImplemented, when the comparison for the given type
                 combination can't be performed.
        :rtype: :py:type:`NotImplementedType`
        """
        try:
            # This routine's purpose is to access the protected method
            # _cmpkey() from client classes, therefore: pylint:
            # disable=protected-access
            return method(self._cmpkey(), other._cmpkey())
        except (AttributeError, TypeError):
            # _cmpkey not implemented, or return different type,
            # so I can't compare with "other".
            # According to the Python Language Reference Manual
            # (http://www.network-theory.co.uk/docs/pylang/Coercionrules.html)
            # return NotImplemented
            return NotImplemented

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)


class Base(ComparableMixin):
    ''' Base class for Fortran 2003 syntax rules.

    All Base classes have the following attributes:
      self.string - original argument to construct a class instance, its type \
                    is either str or FortranReaderBase.
      self.item   - Line instance (holds label) or None.

    :param type cls: the class of object to create.
    :param string: (source of) Fortran string to parse.
    :type string: `str` or \
                  :py:class:`fparser.common.readfortran.FortranReaderBase`
    :param parent_cls: the parent class of this object.
    :type parent_cls: `type`

    '''
    # This dict of subclasses is populated dynamically by code at the end
    # of this module. That code uses the entries in the
    # 'subclass_names' list belonging to each class defined in this module.
    subclasses = {}

    def __init__(self, string, parent_cls=None):
        # pylint:disable=unused-argument
        self.parent = None

    @show_result
    def __new__(cls, string, parent_cls=None):

        def _set_parent(parent_node, items):
            ''' Recursively set the parent of all of the elements
            in the list that are a sub-class of Base. (Recursive because
            sometimes the list of elements itself contains a list or tuple.)

            :param parent_node: the parent of the nodes listed in `items`.
            :type parent_node: sub-class of :py:class:`fparser.two.utils.Base`
            :param items: list or tuple of nodes for which to set the parent.
            :type items: list or tuple of :py:class:`fparser.two.utils.Base` \
                         or `str` or `list` or `tuple` or NoneType.
            '''
            for item in items:
                if item:
                    if isinstance(item, Base):
                        # We can only set the parent of `Base` objects.
                        # Anything else (e.g. str) is passed over.
                        item.parent = parent_node
                    elif isinstance(item, (list, tuple)):
                        _set_parent(parent_node, item)
        # ------------------------------------------------------------------

        from fparser.common import readfortran
        if parent_cls is None:
            parent_cls = [cls]
        elif cls not in parent_cls:
            parent_cls.append(cls)

        # Get the class' match method if it has one
        match = cls.__dict__.get('match')

        if isinstance(string, FortranReaderBase) and \
           match and not issubclass(cls, BlockBase):
            reader = string
            item = reader.get_item()
            if item is None:
                return None
            if isinstance(item, readfortran.Comment):
                # We got a comment but we weren't after a comment (we handle
                # those in Comment.__new__)
                obj = None
            else:
                try:
                    obj = item.parse_line(cls, parent_cls)
                except NoMatchError:
                    obj = None
            if obj is None:
                # No match so give the item back to the reader
                reader.put_item(item)
                return None
            obj.item = item
            return obj

        result = None
        if match:
            # IMPORTANT: if string is FortranReaderBase then cls must
            # restore readers content when no match is found.
            try:
                result = cls.match(string)
            except NoMatchError as msg:
                if str(msg) == '%s: %r' % (cls.__name__, string):
                    # avoid recursion 1.
                    raise

        if isinstance(result, tuple):
            obj = object.__new__(cls)
            obj.string = string
            obj.item = None
            # Set-up parent information for the results of the match
            _set_parent(obj, result)
            if hasattr(cls, 'init'):
                obj.init(*result)
            return obj
        elif isinstance(result, Base):
            return result
        elif result is None:
            # Loop over the possible sub-classes of this class and
            # check for matches
            for subcls in Base.subclasses.get(cls.__name__, []):
                if subcls in parent_cls:  # avoid recursion 2.
                    continue
                try:
                    obj = subcls(string, parent_cls=parent_cls)
                except NoMatchError as msg:
                    obj = None
                if obj is not None:
                    return obj
        else:
            raise AssertionError(repr(result))
        # If we get to here then we've failed to match the current line
        if isinstance(string, FortranReaderBase):
            content = False
            for index in range(string.linecount):
                # Check all lines up to this one for content. We
                # should be able to only check the current line but
                # but as the line number returned is not always
                # correct (due to coding errors) we cannot assume the
                # line pointed to is the line where the error actually
                # happened.
                if string.source_lines[index].strip():
                    content = True
                    break
            if not content:
                # There are no lines in the input or all lines up to
                # this one are empty or contain only white space. This
                # is typically accepted by fortran compilers so we
                # follow their lead and do not raise an exception.
                return
            line = string.source_lines[string.linecount-1]
            errmsg = u"at line {0}\n>>>{1}\n".format(
                string.linecount, line)
        else:
            errmsg = u"{0}: '{1}'".format(cls.__name__, string)
        raise NoMatchError(errmsg)

    def get_root(self):
        '''
        Gets the node at the root of the parse tree to which this node belongs.

        :returns: the node at the root of the parse tree.
        :rtype: :py:class:`fparser.two.utils.Base`

        '''
        current = self
        while current.parent:
            current = current.parent
        return current

    @property
    def children(self):
        '''Return an iterable containing the immediate children of this node in
        the parse tree.

        If this node represents an expression then its children are
        contained in a tuple which is immutable. Therefore, the
        manipulation of the children of such a node must be done by
        replacing the `items` property of the node directly rather than via the
        objects returned by this method.

        :returns: the immediate children of this node.
        :rtype: list or tuple containing zero or more of \
                :py:class:`fparser.two.utils.Base` or NoneType or str

        '''
        child_list = getattr(self, 'content', None)
        if child_list is None:
            child_list = getattr(self, 'items', [])
        return child_list

    def init(self, *items):
        '''
        Store the supplied list of nodes in the `items` list of this node.

        :param items: the children of this node.
        :type items: tuple of :py:class:`fparser.two.utils.Base`

        '''
        self.items = items

    def torepr(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join(map(repr,
                                                                  self.items)))

    def __str__(self):
        return self.tostr()

    def __repr__(self):
        return self.torepr()

    def _cmpkey(self):
        """ Provides a key of objects to be used for comparing.
        """
        return self.items

    def tofortran(self, tab='', isfix=None):
        '''
        Produce the Fortran representation of this Comment.

        :param str tab: characters to pre-pend to output.
        :param bool isfix: whether or not this is fixed-format code.

        :returns: Fortran representation of this comment.
        :rtype: str
        '''
        this_str = six.text_type(self)
        if this_str.strip():
            return tab + this_str
        else:
            # If this_str is empty (i.e this Comment is a blank line) then
            # don't prepend any spaces to it
            return this_str

    def restore_reader(self, reader):
        reader.put_item(self.item)


class BlockBase(Base):
    """
::
    <block-base> = [ <startcls> ]
                     [ <subcls> ]...
                     ...
                     [ <subcls> ]...
                     [ <endcls> ]

content : tuple
    """
    @staticmethod
    def match(startcls, subclasses, endcls, reader,
              match_labels=False,
              match_names=False,
              match_name_classes=(),
              enable_do_label_construct_hook=False,
              enable_if_construct_hook=False,
              enable_where_construct_hook=False,
              enable_select_type_construct_hook=False,
              enable_case_construct_hook=False,
              strict_order=False):
        '''
        Checks whether the content in reader matches the given
        type of block statement (e.g. DO..END DO, IF...END IF etc.)

        :param type startcls: The class marking the beginning of the block
        :param list subclasses: List of classes that can be children of
                                the block
        :param type endcls: The class marking the end of the block
        :param reader: Content to check for match
        :type reader: str or instance of :py:class:`FortranReaderBase`
        :param bool match_labels: TBD
        :param bool match_names: TBD
        :param tuple match_name_classes: TBD
        :param bool enable_do_label_construct_hook: TBD
        :param bool enable_if_construct_hook: TBD
        :param bool enable_where_construct_hook: TBD
        :param bool enable_select_type_construct_hook: TBD
        :param bool enable_case_construct_hook: TBD
        :param bool strict_order: Whether to enforce the order of the
                                  given subclasses.

        :return: instance of startcls or None if no match is found
        :rtype: startcls
        '''
        from fparser.two.Fortran2003 import Comment, Include_Stmt, \
            add_comments_includes_directives
        from fparser.two import C99Preprocessor
        assert isinstance(reader, FortranReaderBase), repr(reader)
        content = []

        if startcls is not None:
            # Deal with any preceding comments, includes, and/or directives
            add_comments_includes_directives(content, reader)
            # Now attempt to match the start of the block
            try:
                obj = startcls(reader)
            except NoMatchError:
                obj = None
            if obj is None:
                # Ultimately we failed to find a match for the
                # start of the block so put back any comments that
                # we processed along the way
                for obj in reversed(content):
                    obj.restore_reader(reader)
                return
            # Store the index of the start of this block proper (i.e.
            # excluding any comments)
            start_idx = len(content)
            content.append(obj)

            if enable_do_label_construct_hook:
                start_label = obj.get_start_label()
            if match_names:
                start_name = obj.get_start_name()

        # Comments and Include statements are always valid sub-classes
        classes = subclasses + [Comment, Include_Stmt]
        # Preprocessor directives are always valid sub-classes
        cpp_classes = [getattr(C99Preprocessor, cls_name)
                       for cls_name in C99Preprocessor.CPP_CLASS_NAMES]
        classes += cpp_classes
        if endcls is not None:
            classes += [endcls]
            endcls_all = tuple([endcls]+endcls.subclasses[endcls.__name__])

        # Start trying to match the various subclasses, starting from
        # the beginning of the list (where else?)
        i = 0
        had_match = False
        found_end = False
        while i < len(classes):
            if enable_do_label_construct_hook:
                try:
                    obj = startcls(reader)
                except NoMatchError:
                    obj = None
                if obj is not None:
                    if start_label == obj.get_start_label():
                        content.append(obj)
                        continue
                    else:
                        obj.restore_reader(reader)
            # Attempt to match the i'th subclass
            cls = classes[i]
            try:
                obj = cls(reader)
            except NoMatchError:
                obj = None
            if obj is None:
                # No match for this class, continue checking the list
                # starting from the i+1'th...
                i += 1
                continue

            # We got a match for this class
            had_match = True
            content.append(obj)

            if match_names and isinstance(obj, match_name_classes):
                end_name = obj.get_end_name()
                if end_name and not start_name:
                    raise FortranSyntaxError(
                        reader, "Name '{0}' has no corresponding starting "
                        "name".format(end_name))
                if end_name and start_name and \
                   end_name.lower() != start_name.lower():
                    raise FortranSyntaxError(
                        reader, "Expecting name '{0}'".format(start_name))

            if endcls is not None and isinstance(obj, endcls_all):
                if match_labels:
                    start_label, end_label = content[start_idx].\
                                             get_start_label(),\
                                             content[-1].get_end_label()
                    if start_label != end_label:
                        continue
                if match_names:
                    start_name, end_name = content[start_idx].\
                                           get_start_name(), \
                                           content[-1].get_end_name()
                    if end_name and not start_name:
                        raise FortranSyntaxError(
                            reader, "Name '{0}' has no corresponding starting "
                            "name".format(end_name))
                    elif start_name and end_name and (start_name.lower() !=
                                                      end_name.lower()):
                        raise FortranSyntaxError(
                            reader, "Expecting name '{0}'".format(start_name))
                # We've found the enclosing end statement so break out
                found_end = True
                break
            if not strict_order:
                # Return to start of classes list now that we've matched.
                i = 0
            if enable_if_construct_hook:
                from fparser.two.Fortran2003 import Else_If_Stmt, Else_Stmt, \
                    End_If_Stmt
                if isinstance(obj, Else_If_Stmt):
                    # Got an else-if so go back to start of possible
                    # classes to match
                    i = 0
                if isinstance(obj, (Else_Stmt, End_If_Stmt)):
                    # Found end-if
                    enable_if_construct_hook = False
            if enable_where_construct_hook:
                from fparser.two.Fortran2003 import Masked_Elsewhere_Stmt, \
                    Elsewhere_Stmt, End_Where_Stmt
                if isinstance(obj, Masked_Elsewhere_Stmt):
                    i = 0
                if isinstance(obj, (Elsewhere_Stmt, End_Where_Stmt)):
                    enable_where_construct_hook = False
            if enable_select_type_construct_hook:
                from fparser.two.Fortran2003 import Type_Guard_Stmt, \
                    End_Select_Type_Stmt
                if isinstance(obj, Type_Guard_Stmt):
                    i = 1
                if isinstance(obj, End_Select_Type_Stmt):
                    enable_select_type_construct_hook = False
            if enable_case_construct_hook:
                from fparser.two.Fortran2003 import Case_Stmt, \
                    End_Select_Stmt
                if isinstance(obj, Case_Stmt):
                    i = 1
                if isinstance(obj, End_Select_Stmt):
                    enable_case_construct_hook = False
            continue

        if not had_match or endcls and not found_end:
            # We did not get a match from any of the subclasses or
            # failed to find the endcls
            if endcls is not None:
                for obj in reversed(content):
                    obj.restore_reader(reader)
                return

        if not content:
            return
        if startcls is not None and endcls is not None:
            # check names of start and end statements:
            start_stmt = content[start_idx]
            end_stmt = content[-1]
            if isinstance(end_stmt, endcls_all) and \
               hasattr(end_stmt, 'get_name') and \
               hasattr(start_stmt, 'get_name'):
                if end_stmt.get_name() is not None:
                    if start_stmt.get_name().string.lower() != \
                       end_stmt.get_name().string.lower():
                        end_stmt.item.reader.error(
                            'expected <%s-name> is %s but got %s. Ignoring.'
                            % (end_stmt.get_type().lower(),
                               start_stmt.get_name(), end_stmt.get_name()))
        return (content,)

    def init(self, content):
        '''
        Initialise the `content` attribute with the list of child nodes.

        :param content: list of nodes that are children of this one.
        :type content: list of :py:class:`fparser.two.utils.Base` or NoneType

        '''
        self.content = content

    def _cmpkey(self):
        """ Provides a key of objects to be used for comparing.
        """
        return self.content

    def tostr(self):
        return self.tofortran()

    def torepr(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.
                           join(map(repr, self.content)))

    def tofortran(self, tab='', isfix=None):
        '''
        Create a string containing the Fortran representation of this class

        :param str tab: indent to prefix to code.
        :param bool isfix: whether or not to generate fixed-format code.

        :return: Fortran representation of this class.
        :rtype: str
        '''
        mylist = []
        start = self.content[0]
        end = self.content[-1]
        extra_tab = ''
        if isinstance(end, EndStmtBase):
            extra_tab = '  '
        if start is not None:
            mylist.append(start.tofortran(tab=tab, isfix=isfix))
        for item in self.content[1:-1]:
            mylist.append(item.tofortran(tab=tab+extra_tab, isfix=isfix))
        if len(self.content) > 1:
            mylist.append(end.tofortran(tab=tab, isfix=isfix))
        # Ensure all strings in list are encoded consistently
        py2_encode_list_items(mylist)
        return '\n'.join(mylist)

    def restore_reader(self, reader):
        for obj in reversed(self.content):
            obj.restore_reader(reader)


class SequenceBase(Base):
    '''
    Match one or more fparser2 rules separated by a defined separator.

    sequence-base is obj [sep obj ] ...

    '''
    @staticmethod
    def match(separator, subcls, string):
        '''Match one or more 'subcls' fparser2 rules in the string 'string'
        separated by 'separator'.

        :param str separator: the separator used to split the supplied \
        string.
        :param subcls: an fparser2 object representing the rule that \
        should be matched.
        :type subcls: subclass of :py:class:`fparser.two.utils.Base`
        :param str string: the input string to match.

        :returns: a tuple containing 1) the separator and 2) the \
        matched objects in a tuple, or None if there is no match.
        :rtype: (str, (Subclass of \
        :py:class:`fparser.two.utils.Base`)) or NoneType

        :raises InternalError: if the separator or string arguments \
        are not the expected type.
        :raises InternalError: if the separator is white space.

        '''
        if not isinstance(separator, (str, six.text_type)):
            raise InternalError(
                "SequenceBase class match method argument separator expected "
                "to be a string but found '{0}'.".format(type(string)))
        if not isinstance(string, (str, six.text_type)):
            raise InternalError(
                "SequenceBase class match method argument string expected to "
                "be a string but found '{0}'.".format(type(string)))

        if separator == ' ':
            raise InternalError(
                "SequenceBase class match method argument separator cannot "
                "be white space.")
        line, repmap = string_replace_map(string)
        splitted = line.split(separator)
        if not splitted:
            # There should be at least one entry.
            return None
        lst = [subcls(repmap(entry.strip())) for entry in splitted]
        return separator, tuple(lst)

    def init(self, separator, items):
        '''Store the result of the match method if the match is successful.

        :param str separator: the separator used to split the supplied string.
        :param items: a tuple containing the matched objects.
        :type items: tuple(Subclass of :py:class:`fparser.two.utils.Base`)

        '''
        self.separator = separator
        self.items = items

    def tostr(self):
        '''
        :returns: The Fortran representation of this object as a string.
        :rtype: str

        '''
        sep = self.separator
        if sep == ',':
            sep = sep + ' '
        elif sep == ' ':
            pass
        else:
            sep = ' ' + sep + ' '
        return sep.join(map(str, self.items))

    def torepr(self):
        '''
        :returns: The Python representation of this object as a string.
        :rtype: str

        '''
        return "{0}('{1}', {2})".format(self.__class__.__name__,
                                        self.separator, self.items)

    # The mixin class is likely to be removed so _cmpkey would not be
    # needed. It is not used at the moment. It is only commented out
    # at this point, rather than removed, in case it turns out that
    # the mixin class is useful.
    # def _cmpkey(self):
    #     """ Provides a key of objects to be used for comparing.
    #     """
    #     return (self.separator, self.items)


class UnaryOpBase(Base):
    """
::
    <unary-op-base> = <unary-op> <rhs>
    """
    def tostr(self):
        return '%s %s' % tuple(self.items)

    def match(op_pattern, rhs_cls, string, exclude_op_pattern=None):
        m = op_pattern.match(string)
        if not m:
            return
        rhs = string[m.end():].lstrip()
        if not rhs:
            return
        op = string[:m.end()].rstrip().upper()
        if exclude_op_pattern is not None:
            if exclude_op_pattern.match(op):
                return
        return op, rhs_cls(rhs)
    match = staticmethod(match)


class BinaryOpBase(Base):
    """
::
    <binary-op-base> = <lhs> <op> <rhs>
    <op> is searched from right by default.
    """
    def match(lhs_cls, op_pattern, rhs_cls, string, right=True,
              exclude_op_pattern=None, is_add=False):
        line, repmap = string_replace_map(string)
        if isinstance(op_pattern, str):
            if right:
                t = line.rsplit(op_pattern, 1)
            else:
                t = line.split(op_pattern, 1)
            if len(t) != 2:
                return
            lhs, rhs = t[0].rstrip(), t[1].lstrip()
            op = op_pattern
        else:
            if right:
                t = op_pattern.rsplit(line, is_add=is_add)
            else:
                t = op_pattern.lsplit(line)
            if t is None or len(t) != 3:
                return
            lhs, op, rhs = t
            lhs = lhs.rstrip()
            rhs = rhs.lstrip()
            op = op.upper()
        if not lhs:
            return
        if not rhs:
            return
        if exclude_op_pattern is not None:
            if exclude_op_pattern.match(op):
                return

        lhs_obj = lhs_cls(repmap(lhs))
        rhs_obj = rhs_cls(repmap(rhs))
        return lhs_obj, op.replace(' ', ''), rhs_obj
    match = staticmethod(match)

    def tostr(self):
        return '%s %s %s' % tuple(self.items)


class SeparatorBase(Base):
    """
::
    <separator-base> = [ <lhs> ] : [ <rhs> ]
    """
    def match(lhs_cls, rhs_cls, string, require_lhs=False, require_rhs=False):
        line, repmap = string_replace_map(string)
        if ':' not in line:
            return
        lhs, rhs = line.split(':', 1)
        lhs = lhs.rstrip()
        rhs = rhs.lstrip()
        lhs_obj, rhs_obj = None, None
        if lhs:
            if lhs_cls is None:
                return
            lhs_obj = lhs_cls(repmap(lhs))
        elif require_lhs:
            return
        if rhs:
            if rhs_cls is None:
                return
            rhs_obj = rhs_cls(repmap(rhs))
        elif require_rhs:
            return
        return lhs_obj, rhs_obj
    match = staticmethod(match)

    def tostr(self):
        s = ''
        if self.items[0] is not None:
            s += '%s :' % (self.items[0])
        else:
            s += ':'
        if self.items[1] is not None:
            s += ' %s' % (self.items[1])
        return s


class KeywordValueBase(Base):
    """
::
    <keyword-value-base> = [ <lhs> = ] <rhs>
    """
    @staticmethod
    def match(lhs_cls, rhs_cls, string, require_lhs=True, upper_lhs=False):
        '''
        :param lhs_cls: list, tuple or single value of classes to attempt to
                        match LHS against (in order), or string containing
                        keyword to match
        :type lhs_cls: names of classes deriving from `:py:class:Base` or str
        :param rhs_cls: name of class to match RHS against
        :type rhs_cls: name of a class deriving from `:py:class:Base`
        :param str string: text to be matched
        :param bool require_lhs: whether the expression to be matched must
                                 contain a LHS that is assigned to
        :param bool upper_lhs: whether or not to convert the LHS of the
                               matched expression to upper case
        :return: instances of the classes representing quantities on the LHS
                 and RHS (LHS is optional) or nothing if no match is found
        :rtype: 2-tuple of objects or nothing
        '''
        if require_lhs and '=' not in string:
            return
        if isinstance(lhs_cls, (list, tuple)):
            for s in lhs_cls:
                obj = KeywordValueBase.match(s, rhs_cls, string,
                                             require_lhs=require_lhs,
                                             upper_lhs=upper_lhs)
                if obj:
                    return obj
            return obj
        # We can't just blindly check whether 'string' contains an '='
        # character as it could itself hold a string constant containing
        # an '=', e.g. FMT='("Hello = False")'
        from fparser.two.Fortran2003 import Char_Literal_Constant
        if not Char_Literal_Constant.match(string) and "=" in string:
            lhs, rhs = string.split('=', 1)
            lhs = lhs.rstrip()
        else:
            lhs = None
            rhs = string.rstrip()
        rhs = rhs.lstrip()
        if not rhs:
            return
        if not lhs:
            if require_lhs:
                return
            return None, rhs_cls(rhs)
        if isinstance(lhs_cls, str):
            if upper_lhs:
                lhs = lhs.upper()
            if lhs_cls != lhs:
                return
            return lhs, rhs_cls(rhs)
        return lhs_cls(lhs), rhs_cls(rhs)

    def tostr(self):
        if self.items[0] is None:
            return str(self.items[1])
        return '%s = %s' % tuple(self.items)


class BracketBase(Base):
    '''
    bracket-base is left-bracket something right-bracket.

    This class is able to cope with nested brackets as long as they
    are correctly nested. Brackets in strings are ignored.

    The 'something' can be specified as being optional.

    '''
    @staticmethod
    def match(brackets, cls, string, require_cls=True):
        '''A generic match method for all types of bracketed
        expressions.

        :param str brackets: the format of the left and right brackets \
        provided as a string, for example '()'
        :param cls: the class to match the content within the brackets \
        :type cls: subclass of :py:class:`fparser.two.utils.Base`
        :param str string: the content to match
        :param bool require_cls: whether the class and associated \
        content is mandatory (True) or optional (False). The default \
        is True.
        :return: None if there is no match, otherwise a tuple with the \
        first and third entries being strings containing the left and \
        right brackets respectively and the second entry being either \
        None or an instance of the class provided as the second \
        argument (cls).
        :rtype: 'NoneType', ( `str`, `NoneType`, `str`) or ( `str`, \
        `cls`, `str` )

        '''
        if not cls and require_cls:
            return None
        if not string:
            return None
        string_strip = string.strip()
        if not brackets:
            return None
        brackets_nospc = brackets.replace(' ', '')
        if not brackets_nospc:
            return None
        if len(brackets_nospc) % 2 == 1:
            # LHS and RHS bracketing must be the same size
            return None
        bracket_len = len(brackets_nospc)//2
        left = brackets_nospc[:bracket_len]
        right = brackets_nospc[-bracket_len:]
        if len(string_strip) < bracket_len*2:
            return None
        if not (string_strip.startswith(left) and
                string_strip.endswith(right)):
            return None
        # Check whether or not there's anything between the open
        # and close brackets
        line = string_strip[bracket_len:-bracket_len].strip()
        if (not line and cls and require_cls) or (line and not cls):
            return None
        if not line and (not cls or not require_cls):
            return left, None, right
        return left, cls(line), right

    def tostr(self):
        '''
        :raises InternalError: if the internal items list variable is \
        not the expected size.
        :raises InternalError: if the first element of the internal \
        items list is None or is an empty string.
        '''

        if len(self.items) != 3:
            raise InternalError(
                "Class BracketBase method tostr() has '{0}' items, "
                "but expecting 3.".format(len(self.items)))
        if not self.items[0]:
            raise InternalError(
                "Class BracketBase method tostr(). 'Items' entry 0 "
                "should be a string containing the left hand bracket "
                "but it is empty or None")
        if not self.items[2]:
            raise InternalError(
                "Class BracketBase method tostr(). 'Items' entry 2 "
                "should be a string containing the right hand bracket "
                "but it is empty or None")
        if self.items[1] is None:
            return "{0}{1}".format(self.items[0], self.items[2])
        return "{0}{1}{2}".format(self.items[0], self.items[1], self.items[2])


class NumberBase(Base):
    """
::
    <number-base> = <number> [ _ <kind-param> ]
    """

    def match(number_pattern, string):
        m = number_pattern.match(string.replace(' ', ''))
        if m is None:
            return
        d = m.groupdict()
        return d['value'].upper(), d.get('kind_param')
    match = staticmethod(match)

    def tostr(self):
        if self.items[1] is None:
            return str(self.items[0])
        return '%s_%s' % tuple(self.items)

    def _cmpkey(self):
        """ Provides a key of objects to be used for comparing.
        """
        return self.items[0]


class CallBase(Base):
    """
::
    <call-base> = <lhs> ( [ <rhs> ] )
    """
    def match(lhs_cls, rhs_cls, string, upper_lhs=False, require_rhs=False):
        if not string.endswith(')'):
            return
        line, repmap = string_replace_map(string)
        i = line.rfind('(')
        if i == -1:
            return
        lhs = line[:i].rstrip()
        if not lhs:
            return
        j = line.rfind(')')
        rhs = line[i+1: j].strip()
        if line[j+1:].lstrip():
            return
        lhs = repmap(lhs)
        if upper_lhs:
            lhs = lhs.upper()
        rhs = repmap(rhs)
        if isinstance(lhs_cls, str):
            if lhs_cls != lhs:
                return
        else:
            lhs = lhs_cls(lhs)
        if rhs:
            if isinstance(rhs_cls, str):
                if rhs_cls != rhs:
                    return
            else:
                rhs = rhs_cls(rhs)
            return lhs, rhs
        elif require_rhs:
            return
        return lhs, None
    match = staticmethod(match)

    def tostr(self):
        if self.items[1] is None:
            return '%s()' % (self.items[0])
        return '%s(%s)' % (self.items[0], self.items[1])


class CALLBase(CallBase):
    """
::
    <CALL-base> = <LHS> ( [ <rhs> ] )
    """
    def match(lhs_cls, rhs_cls, string, require_rhs=False):
        return CallBase.match(lhs_cls, rhs_cls, string,
                              upper_lhs=True, require_rhs=require_rhs)
    match = staticmethod(match)


class StringBase(Base):
    """
::
    <string-base> = <xyz>

Attributes
----------
string
    """
    def match(pattern, string):
        if isinstance(pattern, (list, tuple)):
            for p in pattern:
                obj = StringBase.match(p, string)
                if obj is not None:
                    return obj
            return
        if isinstance(pattern, str):
            if len(pattern) == len(string) and pattern == string:
                return string,
            return
        if pattern.match(string):
            return string,
        return
    match = staticmethod(match)

    def init(self, string):
        self.string = string
        return

    def tostr(self):
        return str(self.string)

    def torepr(self):
        return '%s(%r)' % (self.__class__.__name__, self.string)

    def _cmpkey(self):
        """ Provides a key of objects to be used for comparing.
        """
        return self.string


class STRINGBase(StringBase):
    '''STRINGBase matches an upper case version of the input string with
    another a pattern (typically taken from pattern_tools.py) and
    returns the string in upper case if there is a match.

    '''

    @staticmethod
    def match(my_pattern, string):
        '''Matches an input string with a specified pattern. Casts the string
        to upper case before performing a match and, if there is a
        match, returns the string in upper case.

        The pattern can be a regular expression, a string, a list or a
        tuple. If the input pattern is a regular expression or a
        string, a direct equivalence is performed. If the input pattern is a
        list or a tuple, then all of the contents of the list
        or tuple are searched for a match (by recursing). The list or tuple may
        contain regular expressions, strings, lists or tuples. This
        functionality can be used to recurse down a tree of lists and
        or tuples until regular expressions or strings are found (at
        the leaves of the tree) on which to match. The patterns used
        to match in fparser can be found in patterns_tools.py. These
        make use of the pattern class, whose match method behaves like
        a regular expression. For example:

        from fparser.two import pattern_tools
        pattern = pattern_tools.intrinsic_type_name
        result = STRINGBase.match(pattern, "logical")

        :param pattern: the pattern to match
        :type pattern: `list`, `tuple`, `str` or an `re` expression
        :param str string: the string to match with the pattern
        :return: None if there is no match, or a tuple containing the \
        matched string in upper case.
        :rtype: `NoneType` or ( `str` )

        '''
        if string is None:
            return None
        if not isinstance(string, (str, six.text_type)):
            raise InternalError(
                "Supplied string should be of type str or {0}, but found "
                "{1}".format(six.text_type, type(string)))
        if isinstance(my_pattern, (list, tuple)):
            for child in my_pattern:
                result = STRINGBase.match(child, string)
                if result:
                    return result
            return None
        string_upper = string.upper()
        if isinstance(my_pattern, str):
            if len(my_pattern) == len(string) and my_pattern == string_upper:
                return string_upper,
            return None
        try:
            if my_pattern.match(string_upper):
                return string_upper,
        except AttributeError:
            raise InternalError(
                "Supplied pattern should be a list, tuple, str or regular "
                "expression but found {0}".format(type(my_pattern)))
        return None


class StmtBase(Base):
    """
::
    [ [ <label> ] [ <construct-name> : ] ] <stmt>

Attributes
----------
item : readfortran.Line
    """
    def tofortran(self, tab='', isfix=None):
        label = None
        name = None
        if self.item is not None:
            label = self.item.label
            name = self.item.name
        if isfix:
            c = ' '
        else:
            c = ''
        if label:
            t = c + str(label)
            if isfix:
                while len(t) < 6:
                    t += ' '
            else:
                tab = tab[len(t):] or ' '
        else:
            # BUG allow for fixed format here
            t = ''
        if name:
            return t + tab + name+':' + str(self)
        return t + tab + str(self)

    def get_end_label(self):
        return self.item.label


class EndStmtBase(StmtBase):
    """
::
    <end-stmt-base> = END [ <stmt> [ <stmt-name>] ]
    """
    @staticmethod
    def match(stmt_type, stmt_name, string, require_stmt_type=False):
        start = string[:3].upper()
        if start != 'END':
            return
        line = string[3:].lstrip()
        start = line[:len(stmt_type)].upper()
        if start:
            if start.replace(' ', '') != stmt_type.replace(' ', ''):
                return
            line = line[len(stmt_type):].lstrip()
        else:
            if require_stmt_type:
                return
            return None, None
        if line:
            if stmt_name is None:
                return
            return stmt_type, stmt_name(line)
        return stmt_type, None

    def init(self, stmt_type, stmt_name):
        '''
        Initialise this EndStmtBase object.

        :param str stmt_type: the type of statement, e.g. 'PROGRAM'.
        :param stmt_name: the name associated with the statement or None.
        :type stmt_name: :py:class:`fparser.two.Fortran2003.Name`

        '''
        self.items = [stmt_type, stmt_name]

    def get_name(self):
        return self.items[1]

    def get_type(self):
        return self.items[0]

    def tostr(self):
        if self.items[1] is not None:
            return 'END %s %s' % tuple(self.items)
        if self.items[0] is not None:
            return 'END %s' % (self.items[0])
        return 'END'

    def torepr(self):
        return '%s(%r, %r)' % (
            self.__class__.__name__, self.get_type(), self.get_name())

    def get_end_name(self):
        name = self.items[1]
        if name is not None:
            return name.string


def isalnum(c):
    return c.isalnum() or c == '_'


class WORDClsBase(Base):
    '''Base class to support situations where there is a keyword which is
    optionally followed by further text, potentially separated by
    '::'.

    For example 'program fred', or 'import :: a,b'

    WORD-cls is WORD [ [ :: ] cls ]

    '''
    @staticmethod
    def match(keyword, cls, string, colons=False, require_cls=False):
        '''Checks whether the content in string matches the expected
        WORDClsBase format with 'keyword' providing the keyword, 'cls'
        providing the following text, 'colons' specifying whether an
        optional '::' is allowed as a separator between the keyword
        and cls and 'require_cls' specifying whether cls must have
        content or not.

        Note, if the optional '::' is allowed and exists in the string
        then 1) cls must also have content i.e. it implies
        `require_cls=True` and 2) white space is not required between
        the keyword and the '::' and the '::' and cls.

        The simplest form of keyword pattern is a string. However this
        method can also match more complex patterns as specified by
        the Pattern class in pattern_tools.py. As patterns can be
        built from combinations of other patterns (again see
        pattern_tool.py) this method also supports a hierarchy of
        lists and/or tuples of patterns.

        :param keyword: the pattern of the WORD to match. This can be \
            a Pattern, string, list or tuple, with a list or tuple \
            containing one or more Pattern, string, list or tuple.
        :type keyword: :py:class:`fparser.two.pattern_tools.Pattern`, \
            str, tuple of str/Pattern/tuple/list or list of \
            str/Pattern/tuple/list
        :param cls: the class to match.
        :type cls: a subclass of :py:class:`fparser.two.utils.Base`
        :param str string: Text that we are trying to match.
        :param bool colons: whether '::' is allowed as an optional \
            separator between between WORD and cls.
        :param bool require_cls: whether content for cls is required \
            or not.

        :returns: None if there is no match or, if there is a match, a \
            2-tuple containing a string matching the 'WORD' and an \
            instance of 'cls' (or None if an instance of cls is not \
            required and not provided).
        :rtype: (str, cls or NoneType) or NoneType

        '''
        if isinstance(keyword, (tuple, list)):
            for child in keyword:
                try:
                    obj = WORDClsBase.match(child, cls, string,
                                            colons=colons,
                                            require_cls=require_cls)
                except NoMatchError:
                    obj = None
                if obj is not None:
                    return obj
            return None

        if isinstance(keyword, str):
            line = string.lstrip()
            if line[:len(keyword)].upper() != keyword.upper():
                return None
            line = line[len(keyword):]
            pattern_value = keyword
        else:
            my_match = keyword.match(string)
            if my_match is None:
                return None
            line = string[len(my_match.group()):]
            pattern_value = keyword.value

        if not line:
            if require_cls:
                # no text found but it is required
                return None
            return pattern_value, None
        if isalnum(line[0]):
            return None
        line = line.lstrip()
        has_colons = False
        if colons and line.startswith('::'):
            has_colons = True
            line = line[2:].lstrip()
        if not line:
            if has_colons or require_cls:
                # colons without following content is not allowed.
                return None
            return pattern_value, None
        if cls is None:
            return None
        return pattern_value, cls(line)

    def tostr(self):
        '''Convert the class into Fortran.

        :return: String representation of this class without any \
                 optional '::'.
        :rtype: str

        '''
        if self.items[1] is None:
            return str(self.items[0])
        s = str(self.items[1])
        if s and s[0] in '(*':
            return '%s%s' % (self.items[0], s)
        return '%s %s' % (self.items[0], s)

    def tostr_a(self):
        '''Convert the class into Fortran, adding in "::".

        :return: String representation of this class including an \
                 optional '::'.
        :rtype: str

        '''
        if self.items[1] is None:
            return str(self.items[0])
        return '%s :: %s' % (self.items[0], self.items[1])


class Type_Declaration_StmtBase(StmtBase):
    """<type-declaration-stmt> = <declaration-type-spec> [ [ ,
    <attr-spec> ]... :: ] <entity-decl-list>

    """
    subclass_names = []
    use_names = None  # derived class must define this list

    @staticmethod
    def match(decl_type_spec_cls, attr_spec_list_cls,
              entity_decl_list_cls, string):
        line, repmap = string_replace_map(string)
        i = line.find('::')
        if i != -1:
            j = line[:i].find(',')
            if j != -1:
                i = j
        else:
            if line[:6].upper() == 'DOUBLE':
                m = re.search(r'\s[a-z_]', line[6:].lstrip(), re.I)
                if m is None:
                    return
                i = m.start() + len(line)-len(line[6:].lstrip())
            else:
                m = re.search(r'\s[a-z_]', line, re.I)
                if m is None:
                    return
                i = m.start()
        type_spec = decl_type_spec_cls(repmap(line[:i].rstrip()))
        if type_spec is None:
            return
        line = line[i:].lstrip()
        if line.startswith(','):
            i = line.find('::')
            if i == -1:
                return
            attr_specs = attr_spec_list_cls(repmap(line[1:i].strip()))
            if attr_specs is None:
                return
            line = line[i:]
        else:
            attr_specs = None
        if line.startswith('::'):
            line = line[2:].lstrip()
        entity_decls = entity_decl_list_cls(repmap(line))
        if entity_decls is None:
            return
        return type_spec, attr_specs, entity_decls

    def tostr(self):
        if self.items[1] is None:
            return '%s :: %s' % (self.items[0], self.items[2])
        else:
            return '%s, %s :: %s' % self.items


def walk(node_list, types=None, indent=0, debug=False):
    '''
    Walk down the parse tree produced by fparser2.  Returns a list of all
    nodes with the specified type(s).

    :param node_list: node or list of nodes from which to walk.
    :type node_list: (list of) :py:class:fparser.two.utils.Base
    :param types: type or tuple of types of Node to return. (Default is to \
                  return all nodes.)
    :type types: type or tuple of types
    :param int indent: extent to which to indent debug output.
    :param bool debug: whether or not to write textual representation of AST \
                       to stdout.
    :returns: a list of nodes
    :rtype: `list` of :py:class:`fparser.two.utils.Base`
    '''
    local_list = []

    if not isinstance(node_list, (list, tuple)):
        node_list = [node_list]

    for child in node_list:
        if debug:
            if isinstance(child, str):
                print(indent*"  " + "child type = ", type(child), repr(child))
            else:
                print(indent*"  " + "child type = ", type(child))
        if types is None or isinstance(child, types):
            local_list.append(child)
        # Recurse down
        if isinstance(child, Base):
            local_list += walk(child.children, types, indent+1, debug)

    return local_list


def get_child(node, node_type):
    '''
    Searches for the first, immediate child of the supplied node that is of
    the specified type.

    :param node: the node whose children will be searched.
    :type node: :py:class:`fparser.two.utils.Base`
    :param type node_type: the class of child node to search for.

    :returns: the first child node of type node_type that is encountered \
              or None.
    :rtype: :py:class:`fparser.two.utils.Base`

    '''
    for child in node.children:
        if isinstance(child, node_type):
            return child
    return None



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2019-2020, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

'''Test that Fortran 2003 intrinsic functions are parsed correctly.'''

import pytest
from fparser.two.Fortran2003 import Program, Intrinsic_Function_Reference, \
    Intrinsic_Name
from fparser.two.utils import FortranSyntaxError, NoMatchError, \
    InternalSyntaxError, walk
from fparser.api import get_reader

# class program


@pytest.mark.usefixtures("f2003_create")
def test_intrinsic_recognised():
    '''Test that an intrinsic is picked up when used in a program.'''

    reader = get_reader("subroutine sub()\na = sin(b)\nend subroutine sub\n")
    ast = Program(reader)
    assert walk(ast, Intrinsic_Function_Reference)


def test_intrinsic_error(f2003_create):
    '''Test that Program raises the expected exception when there is an
    intrinsic syntax error.

    '''
    reader = get_reader("subroutine sub()\na = sin(b,c)\nend subroutine sub\n")
    with pytest.raises(FortranSyntaxError) as excinfo:
        _ = Program(reader)
    assert (
        "at line 2\n"
        ">>>a = sin(b,c)\n"
        "Intrinsic 'SIN' expects 1 arg(s) but found 2") in str(excinfo.value)

# class intrinsic_name


def test_intrinsic_name_generic(f2003_create):
    '''Test that class Intrinsic_Name correctly matches a generic name.'''
    result = Intrinsic_Name("COS")
    assert isinstance(result, Intrinsic_Name)
    assert str(result) == "COS"


def test_intrinsic_name_specific(f2003_create):
    '''Test that class Intrinsic_Name correctly matches a specific name.'''
    result = Intrinsic_Name("CCOS")
    assert isinstance(result, Intrinsic_Name)
    assert str(result) == "CCOS"


def test_intrinsic_name_invalid(f2003_create):
    '''Test that class Intrinsic_Name raises the expected exception if an
    invalid intrinsic name is provided.

    '''
    with pytest.raises(NoMatchError):
        _ = Intrinsic_Name("NOT_AN_INTRINSIC")


def test_intrinsic_name_case_insensitive(f2003_create):
    '''Test that class Intrinsic_Name is a case insensitive match which
    returns the name in upper case.

    '''
    result = Intrinsic_Name("CcoS")
    assert isinstance(result, Intrinsic_Name)
    assert str(result) == "CCOS"

# class intrinsic_function_reference


def test_intrinsic_function_reference_generic(f2003_create):
    '''Test that class Intrinsic_Function_Reference correctly matches a
    generic intrinsic with a valid number of arguments.

    '''
    result = Intrinsic_Function_Reference("SIN(A)")
    assert isinstance(result, Intrinsic_Function_Reference)
    assert str(result) == "SIN(A)"


def test_intrinsic_function_reference(f2003_create):
    '''Test that class Intrinsic_Function_Reference correctly matches a
    specific intrinsic with a valid number of arguments.

    '''
    result = Intrinsic_Function_Reference("DSIN(A)")
    assert isinstance(result, Intrinsic_Function_Reference)
    assert str(result) == "DSIN(A)"


def test_intrinsic_function_nomatch(f2003_create):
    '''Test that class Intrinsic_Function_Reference raises the expected
    exception if there is no match.

    '''
    with pytest.raises(NoMatchError):
        _ = Intrinsic_Function_Reference("NO_MATCH(A)")


def test_intrinsic_function_reference_multi_args(f2003_create):
    '''Test that class Intrinsic_Function_Reference correctly matches a
    generic intrinsic which accepts more than one argument (two in
    this case).

    '''
    result = Intrinsic_Function_Reference("MATMUL(A,B)")
    assert isinstance(result, Intrinsic_Function_Reference)
    assert str(result) == "MATMUL(A, B)"


def test_intrinsic_function_reference_zero_args(f2003_create):
    '''Test that class Intrinsic_Function_Reference correctly matches a
    generic intrinsic which accepts zero arguments.

    '''
    result = Intrinsic_Function_Reference("COMMAND_ARGUMENT_COUNT()")
    assert isinstance(result, Intrinsic_Function_Reference)
    assert str(result) == "COMMAND_ARGUMENT_COUNT()"


def test_intrinsic_function_reference_range_args(f2003_create):
    '''Test that class Intrinsic_Function_Reference correctly matches a
    generic intrinsic which accepts a range of number of arguments.

    '''
    for args in ["", "A", "A, B", "A, B, C"]:
        result = Intrinsic_Function_Reference("SYSTEM_CLOCK({0})".format(args))
        assert isinstance(result, Intrinsic_Function_Reference)
        assert str(result) == "SYSTEM_CLOCK({0})".format(args)


def test_intrinsic_function_reference_unlimited_args(f2003_create):
    '''Test that class Intrinsic_Function_Reference correctly matches a
    generic intrinsic which accepts an unlimitednumber of arguments.

    '''
    for args in ["A, B", "A, B, C", "A, B, C, D"]:
        result = Intrinsic_Function_Reference("MAX({0})".format(args))
        assert isinstance(result, Intrinsic_Function_Reference)
        assert str(result) == "MAX({0})".format(args)


def test_intrinsic_function_reference_error1(f2003_create):
    '''Test that class Intrinsic_Function_Reference raises the expected
    exception when the valid min and max args are equal (2 in this case)
    and the wrong number of arguments is supplied.

    '''
    with pytest.raises(InternalSyntaxError) as excinfo:
        _ = Intrinsic_Function_Reference("MATMUL(A)")
    assert ("Intrinsic 'MATMUL' expects 2 arg(s) but found 1."
            "" in str(excinfo.value))

    with pytest.raises(InternalSyntaxError) as excinfo:
        _ = Intrinsic_Function_Reference("MATMUL(A,B,C)")
    assert ("Intrinsic 'MATMUL' expects 2 arg(s) but found 3."
            "" in str(excinfo.value))


def test_intrinsic_function_reference_error2(f2003_create):
    '''Test that class Intrinsic_Function_Reference raises the expected
    exception when the valid min args is less than the valid max args
    and the wrong number of arguments is supplied.

    '''
    with pytest.raises(InternalSyntaxError) as excinfo:
        _ = Intrinsic_Function_Reference("PRODUCT()")
    assert ("Intrinsic 'PRODUCT' expects between 1 and 3 args but found 0."
            "" in str(excinfo.value))

    with pytest.raises(InternalSyntaxError) as excinfo:
        _ = Intrinsic_Function_Reference("PRODUCT(A,B,C,D)")
    assert ("Intrinsic 'PRODUCT' expects between 1 and 3 args but found 4."
            "" in str(excinfo.value))


def test_intrinsic_function_reference_error3(f2003_create):
    '''Test that class Intrinsic_Function_Reference raises the expected
    exception when the number of arguments is unlimited.

    '''
    with pytest.raises(InternalSyntaxError) as excinfo:
        _ = Intrinsic_Function_Reference("MIN(A)")
    assert ("Intrinsic 'MIN' expects at least 2 args but found 1."
            "" in str(excinfo.value))


def test_intrinsic_inside_intrinsic(f2003_create):
    '''Test that when an intrinsic is within another instrinsic then both
    are recognised as intrinsics.

    '''
    reader = get_reader("subroutine sub()\na = sin(cos(b))\nend "
                        "subroutine sub\n")
    ast = Program(reader)
    rep = repr(ast).replace("u'", "'")
    assert "Intrinsic_Name('SIN')" in rep
    assert "Intrinsic_Name('COS')" in rep



# Copyright (c) 2019-2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R701 : primary type.

Sub-rules:
    C701 (R701) The type-param-name shall be the name of a type parameter.
    C702 (R701) The designator shall not be a whole assumed-size array

Neither C701 nor C702 can be tested here, as they require context of the
type defined outside of Primary.

'''
import sys

import pytest

import fparser.two.Fortran2003 as f2003
import fparser.two.utils


def assert_subclass_parse(source, base_type, actual_type=None,
                          expected_str=None):
    '''Assert that the given source matches the given ``base_type``
    and optionally the specific type that it should produce.

    :param source: The Fortran source to be parsed.
    :type source: str or :py:class:`FortranReaderBase`
    :param base_type: the base type from which a match is expected to be found
    :type base_type: :py:class:`fortran.two.Fortran2003.Base` subclass
    :param actual_type: The actual type matched by the parser.
    :type actual_type: :py:class:`fortran.two.Fortran2003.Base` subclass
    :param str expected_str: The expected ``str(result)`` of the parsed result

    '''
    obj = f2003.Primary(source)
    # Note: Check that the type exists in the possible types, rather than
    # checking the type is an instance of one of the possible types (and
    # all of its ancestor types). This is a much more targetted test as a
    # result.
    assert type(obj) in possible_subclasses(base_type)

    if actual_type:
        assert isinstance(obj, actual_type)
    else:
        assert isinstance(obj, base_type)

    if expected_str:
        assert expected_str == str(obj)


def possible_subclasses(node_type, _seen=None):
    '''Given a type (e.g. Fortran2003.Primary), return all of the
    subtypes that could have been matched after parsing.

    NOTE: This is not a general implementation. It is useful for testing
    in some limited situations. Please refer to the Base.__new__ for the
    actual logic for identifying possible_subclasses.

    :param node_type: The root node from which to find all subclasses.
    :type node_type: :py:class:`fortran.two.Fortran2003.Base` subclass
    :param _seen: Private list of seen subclasses, designed to support \
                  recursive calls to this function.
    :type _seen: None or list

    '''
    seen = _seen or []
    subclasses = getattr(node_type, 'subclass_names', [])
    if node_type not in seen:
        seen.append(node_type)

    for subclass_name in subclasses:
        module = sys.modules[node_type.__module__]
        subclass = getattr(module, subclass_name, None)
        if subclass is not None and subclass not in seen:
            seen.append(subclass)
            possible_subclasses(subclass, seen)
    return seen


def test_intrinsic_function(f2003_create):
    '''Test that an intrinsic function is matched by Primary.
    '''
    assert_subclass_parse(
        'sin(x)', f2003.Intrinsic_Function_Reference,
        actual_type=f2003.Intrinsic_Function_Reference,
        expected_str='SIN(x)')


def test_constant(f2003_create):
    '''Test that Constant types are matched by Primary.
    '''
    assert_subclass_parse(
        '1.2e-03', f2003.Constant,
        actual_type=f2003.Real_Literal_Constant,
        expected_str='1.2E-03')


def test_designator(f2003_create):
    '''Test that Designator types are matched by Primary.
    '''
    assert_subclass_parse(
        'array(1:5)', f2003.Designator,
        actual_type=f2003.Part_Ref,
        expected_str='array(1 : 5)')


def test_array_constructor(f2003_create):
    '''Test that Array Constructor types are matched by Primary.
    '''
    assert_subclass_parse(
        '[ 1.2, 2.3e + 2,    -5.1 e-3 ]', f2003.Array_Constructor,
        actual_type=f2003.Array_Constructor,
        expected_str='[1.2, 2.3E+2, - 5.1E-3]')


def test_structure_constructor(f2003_create):
    '''Test that Structure Constructor types are matched by Primary.
    '''
    # Note: The actual returned type is Data_Ref. With more context of what
    # has already been parsed it is possible that this could change in the
    # future. For instance, in this example "PERSON" could actually be the
    # name of a function.
    assert_subclass_parse(
        'PERSON ( 12,   "Jones" )', f2003.Structure_Constructor,
        actual_type=f2003.Part_Ref,
        expected_str='PERSON(12, "Jones")')


@pytest.mark.xfail(reason="Requires more parse context (#190)")
def test_function_reference(f2003_create):
    '''This test demonstrates the inability to distinguish
    Structure_Constructor from Function_Reference without more parse context
    than is currently being provided.
    '''
    assert_subclass_parse(
        'a_function(1.2, some_kwarg="hello")', f2003.Function_Reference)


@pytest.mark.xfail(reason="Requires more parse context (#190)")
def test_type_param_inquiry():
    '''This test demonstrates the inability to distinguish Designator from
    Type_Param_Inquiry without more parse context than is currently being
    provided.
    '''
    assert_subclass_parse(
        'X % KIND', f2003.Type_Param_Inquiry)


def test_type_param_name():
    '''Test that Type_Param_Name types are matched by Primary.
    '''
    assert_subclass_parse(
        'INTEGER', f2003.Type_Param_Name,
        actual_type=f2003.Name,
        expected_str='INTEGER')


def test_parenthesis(f2003_create):
    '''Test that Parenthesis types are matched by Primary.
    '''
    assert_subclass_parse(
        '(a +  b)', f2003.Parenthesis,
        expected_str='(a + b)')


def test_no_match(f2003_create):
    '''Test that a NoMatchError is raised if we provide code
    that isn't allowed as a Primary type (e.g. a comment).
    '''
    with pytest.raises(fparser.two.utils.NoMatchError):
        _ = f2003.Primary('! A comment')


@pytest.mark.xfail(reason="Requires more parse context (#190)")
def test_c701_no_assumed_size_array(f2003_create):
    '''Test C701 (R701) The type-param-name shall be the name of a type.
    This test cannot be passed without more parse context of things like
    defined types.
    '''
    context = f2003.Type_Declaration_Stmt("INTEGER :: not_a_type")
    with pytest.raises(fparser.two.utils.NoMatchError):
        f2003.Primary('not_a_type',)  # context)


@pytest.mark.xfail(reason="Requires more parse context (#190)")
def test_c702_no_assumed_size_array(f2003_create):
    '''Test C702 (R701) The designator shall not be a whole assumed-size array.
    This test cannot be passed without more parse context of things like
    defined types.
    '''
    context = f2003.Type_Declaration_Stmt("integer(*) :: assumed_size_array")
    with pytest.raises(fparser.two.utils.NoMatchError):
        f2003.Primary('assumed_size_array',)  # context)



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''File containing unit tests for the STRINGBase baseclass in
utils.py'''

import pytest
from fparser.two.utils import STRINGBase, InternalError


def test_string():
    '''Test the STRINGbase match method with a string pattern.'''

    pattern = "HELLO"
    for my_input in ["hello", "HeLlO", "HELLO"]:
        result = STRINGBase.match(pattern, my_input)
        assert repr(result) == "('{0}',)".format(pattern)
        assert str(result[0]) == pattern


def test_re():
    '''Test the STRINGbase match method with a regular expression.'''

    import re
    pattern = re.compile('[A-Z][0-9]+')
    for my_input in ["a123", "A123"]:
        result = STRINGBase.match(pattern, my_input)
        assert repr(result) == "('{0}',)".format(my_input.upper())
        assert str(result[0]) == my_input.upper()


def test_list():
    '''Test the STRINGbase match method with a list.'''

    import re
    pattern1 = re.compile('[A-Z][0-9]+')
    pattern2 = "HELLO"
    pattern_list = [pattern1, pattern2]
    for my_input in ["a123", "A123", "hello", "HeLlO", "HELLO"]:
        result = STRINGBase.match(pattern_list, my_input)
        assert repr(result) == "('{0}',)".format(my_input.upper())
        assert str(result[0]) == my_input.upper()


def test_tuple():
    '''Test the STRINGbase match method with a tuple.'''

    import re
    pattern1 = re.compile('[A-Z][0-9]+')
    pattern2 = "HELLO"
    pattern_tuple = (pattern1, pattern2)
    for my_input in ["a123", "A123", "hello", "HeLlO", "HELLO"]:
        result = STRINGBase.match(pattern_tuple, my_input)
        assert repr(result) == "('{0}',)".format(my_input.upper())
        assert str(result[0]) == my_input.upper()


def test_pattern_class():
    '''Test the STRINGbase match method with a pattern instance as
    specified in pattern_tools.py.

    '''

    from fparser.two import pattern_tools
    pattern = pattern_tools.intrinsic_type_name
    for my_input in ["logical", "LoGiCaL", "LOGICAL"]:
        result = STRINGBase.match(pattern, my_input)
        assert repr(result) == "('{0}',)".format(my_input.upper())
        assert str(result[0]) == my_input.upper()


def test_invalid_pattern():
    '''Test the STRINGbase match method with an invalid type of
    pattern.

    '''

    for invalid_pattern in [None, 123]:
        with pytest.raises(InternalError) as excinfo:
            _ = STRINGBase.match(invalid_pattern, "hello")
        assert ("Supplied pattern should be a list, tuple, str or regular "
                "expression but found {0}".format(type(invalid_pattern))
                in str(excinfo.value))


def test_None_string():
    '''Test the STRINGbase match method returns None when the string is
    None.

    '''
    result = STRINGBase.match("hello", None)
    assert result is None


def test_not_string():
    '''Test that the STRINGbase match method returns an internal error
    when the string argument is not a string (and is not None), as this
    is not something we would expect and so would indicate that
    something serious has gone wrong.

    '''
    with pytest.raises(InternalError) as excinfo:
        _ = STRINGBase.match("hello", 123)
    assert "Supplied string should be of type str" in str(excinfo.value)



# Copyright (c) 2018-2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R1109 : This file tests the support for the
Use statement.

'''

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Use_Stmt
from fparser.two.utils import NoMatchError, InternalError

# match() use ...


# match() 'use x'. Use both string and reader input here, but from
# here on we will just use string input as that is what is passed to
# the match() method
def test_use(f2003_create):
    '''Check that a basic use is parsed correctly. Input separately as a
    string and as a reader object

    '''
    def check_use(reader):
        '''Internal helper function to avoid code replication.'''
        ast = Use_Stmt(reader)
        assert "USE my_model" in str(ast)
        assert repr(ast) == "Use_Stmt(None, None, Name('my_model'), '', None)"

    line = "use my_model"
    check_use(line)
    reader = get_reader(line)
    check_use(reader)


# match() 'use :: x'
def test_use_colons(f2003_create):
    '''Check that a basic use with '::' is parsed correctly.'''
    line = "use :: my_model"
    ast = Use_Stmt(line)
    assert "USE :: my_model" in str(ast)
    assert repr(ast) == "Use_Stmt(None, '::', Name('my_model'), '', None)"


# match() 'use, nature :: x'
def test_use_nature(f2003_create):
    '''Check that a use with a 'nature' specification is parsed correctly.'''
    line = "use, intrinsic :: my_model"
    ast = Use_Stmt(line)
    assert "USE, INTRINSIC :: my_model" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(Module_Nature('INTRINSIC'), '::', Name('my_model'), "
        "'', None)")


# match() 'use x, rename'
def test_use_rename(f2003_create):
    '''Check that a use with a nename clause is parsed correctly.'''
    line = "use my_module, name=>new_name"
    ast = Use_Stmt(line)
    assert "USE my_module, name => new_name" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_module'), ',', Rename_List(',', "
        "(Rename(None, Name('name'), Name('new_name')),)))")


# match() 'use x, only: y'
def test_use_only(f2003_create):
    '''Check that a use statement is parsed correctly when there is an
    only clause.

    '''
    line = "use my_model, only: name"
    ast = Use_Stmt(line)
    assert "USE my_model, ONLY: name" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_model'), ', ONLY:', Only_List(',', "
        "(Name('name'),)))")


# match() 'use x, only:'
def test_use_only_empty(f2003_create):
    '''Check that a use statement is parsed correctly when there is an
    only clause without any content.

    '''
    line = "use my_model, only:"
    ast = Use_Stmt(line)
    assert "USE my_model, ONLY:" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(None, None, Name('my_model'), ', ONLY:', None)")


# match() '  use  ,  nature  ::  x  ,  name=>new_name'
def test_use_spaces_1(f2003_create):
    '''Check that a use statement with spaces works correctly with
    renaming.

    '''
    line = "  Use  ,  intrinsic  ::  my_model  ,  name=>new_name  "
    ast = Use_Stmt(line)
    assert "USE, INTRINSIC :: my_model, name => new_name" in str(ast)
    assert repr(ast) == (
        "Use_Stmt(Module_Nature('INTRINSIC'), '::', Name('my_model'), ',', "
        "Rename_List(',', (Rename(None, Name('name'), Name('new_name')),)))")


# match() '  use  ,  nature  ::  x  ,  only  :  name'
def test_use_spaces_2(f2003_create):
    '''Check that a use statement with spaces works correctly with an only
    clause.

    '''
    line = "  use  ,  intrinsic  ::  my_model  ,  only  :  name  "
    ast = Use_Stmt(line)
    assert "USE, INTRINSIC :: my_model, ONLY: name" in str(ast)
    assert (repr(ast) ==
            "Use_Stmt(Module_Nature('INTRINSIC'), '::', Name('my_model'), ', "
            "ONLY:', Only_List(',', (Name('name'),)))")


# match() mixed case
def test_use_mixed_case(f2003_create):
    '''Check that a use statement with mixed case keywords ('use' and
    'only') works as expected.

    '''
    line = "UsE my_model, OnLy: name"
    ast = Use_Stmt(line)
    assert "USE my_model, ONLY: name" in str(ast)
    assert (repr(ast) ==
            "Use_Stmt(None, None, Name('my_model'), ', ONLY:', Only_List(',', "
            "(Name('name'),)))")

# match() Syntax errors


def test_syntaxerror(f2003_create):
    '''Test that NoMatchError is raised for various syntax errors.'''
    for line in ["us", "ust", "use", "usemy_model", "use, ", "use, ::",
                 "use, intrinsic", "use, intrinsic::",
                 "use, intrinsic my_module", "use,", "use my_model,",
                 "use my_model, only", "use my_model, only ;",
                 "use my_model, only name"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Use_Stmt(line)
        assert "Use_Stmt: '{0}'".format(line) in str(excinfo.value)

# match() Internal errors


def test_use_internal_error1(f2003_create):
    '''Check that an internal error is raised if the length of the Items
    list is not 5 as the str() method assumes that it is.

    '''
    line = "use my_model"
    ast = Use_Stmt(line)
    ast.items = (None, None, None, None)
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "should be of size 5 but found '4'" in str(excinfo.value)


def test_use_internal_error2(f2003_create):
    '''Check that an internal error is raised if the module name (entry 2
    of Items) is empty or None as the str() method assumes that it is
    a string with content.

    '''
    line = "use my_model"
    ast = Use_Stmt(line)
    for content in [None, ""]:
        ast.items = (None, None, content, None, None)
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("entry 2 should be a module name but it is "
                "empty") in str(excinfo.value)


def test_use_internal_error3(f2003_create):
    '''Check that an internal error is raised if entry 3 of Items is
    'None' as the str() method assumes it is a (potentially empty)
    string.

    '''
    line = "use my_model"
    ast = Use_Stmt(line)
    ast.items = (None, None, "my_module", None, None)
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "entry 3 should be a string but found 'None'" in str(excinfo.value)



# Copyright (c) 2018-2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test utils.py which contain base classes to support fparser,
exception handling and ast traversal.

'''

import pytest
from fparser.two.utils import FortranSyntaxError
from fparser.api import get_reader

# test BlockBase


def test_blockbase_match_names(f2003_create):
    '''Test the blockbase name matching option in its match method. We use
    the Derived_Type_Def class (which subclasses BlockBase) for this
    as it sets match_names to True.

    '''
    from fparser.two.Fortran2003 import Derived_Type_Def, Case_Construct

    # working named example
    reader = get_reader("type abc\nend type abc")
    ast = Derived_Type_Def(reader)
    assert "TYPE :: abc\nEND TYPE abc" in str(ast)

    # case insensitive
    reader = get_reader("type abc\nend type ABC")
    ast = Derived_Type_Def(reader)
    assert "TYPE :: abc\nEND TYPE ABC" in str(ast)

    # incorrect name exception
    reader = get_reader("type abc\nend type cde")
    with pytest.raises(FortranSyntaxError) as excinfo:
        ast = Derived_Type_Def(reader)
    assert "at line 2\n>>>end type cde\nExpecting name 'abc'" \
        in str(excinfo.value)

    # first name required if second name supplied
    # switch to using select case as it can trip the exception
    reader = get_reader("select case (i)\nend select label")
    with pytest.raises(FortranSyntaxError) as excinfo:
        ast = Case_Construct(reader)
    assert ("at line 2\n>>>end select label\nName 'label' has no "
            "corresponding starting name") in str(excinfo.value)


def test_blockbase_match_name_classes(f2003_create):
    '''Test the blockbase name matching option in its match method. We use
    the If_Construct class (which subclasses BlockBase) for this as it
    sets match_names to True and provides match_name_classes. This is
    used when names can appear in multiple places.

    '''
    from fparser.two.Fortran2003 import If_Construct

    # working named example
    reader = get_reader("label:if (.true.) then\nendif label")
    ast = If_Construct(reader)
    assert "label:IF (.TRUE.) THEN\nEND IF label" in str(ast)

    # case insensitive
    reader = get_reader("label:if (.true.) then\nendif LABEL")
    ast = If_Construct(reader)
    assert "label:IF (.TRUE.) THEN\nEND IF LABEL" in str(ast)

    # incorrect name exception
    reader = get_reader("label:if (.true.) then\nendif bella")
    with pytest.raises(FortranSyntaxError) as excinfo:
        ast = If_Construct(reader)
    assert "at line 2\n>>>endif bella\nExpecting name 'label'" \
        in str(excinfo.value)

    # first name required if subsequent name supplied
    reader = get_reader("if (.true.) then\nendif label")
    with pytest.raises(FortranSyntaxError) as excinfo:
        ast = If_Construct(reader)
    assert ("at line 2\n>>>endif label\nName 'label' has no corresponding "
            "starting name") in str(excinfo.value)
